import os
import shutil
from joblib import Parallel, delayed
from .data_file import DataFile
from .helpers import (proc_file_input, mp_consol_save, wrap_load_func)
from pandas.util._decorators import doc
from .Dataset import _shared_docs, _sip_docs
import numpy as np
import pandas as pd


def get_file_mapping(self):
    '''This function is used to access the
    up to date file mapping.

    Returns
    --------
    file_mapping : dict
        Return a dictionary with keys as
        integer's loaded in the Dataset referring
        to Data Files.

    See Also
    --------
    to_data_file : Cast existing columns to type Data File.
    add_data_files : Method for adding new data files

    '''

    self._check_file_mapping()
    return self.file_mapping


@doc(load_func=_shared_docs['load_func'], inplace=_shared_docs['inplace'])
def add_data_files(self, files, file_to_subject,
                   load_func=np.load, inplace=False):
    '''This method allows adding columns of type
    'data file' to the Dataset class.

    Parameters
    ----------
    files : dict
        | This argument specifies the files to be loaded as :ref:`data_files`.
          Files must be passed as a python dict where
          each key refers to the name of that feature / column of data files
          to load, and the value is either a list-like of
          str file paths, or a single globbing str which will
          be used to determine the files.

        | In addition to this parameter, you must also pass a
          python function to the file_to_subject param,
          which specifies how to convert from passed
          file path, to a subject name.

    file_to_subject : python function, dict of or 'auto'
        | This parameter represents how the subject name should
          be determined from the passed file paths. This
          parameter can be passed any python function, where
          the first argument on the function takes a full
          file path and returns a subject name.

        | This parameter should be passed as either a single function
          or argument to be used for all columns, or as a dictionary
          corresponding to the passed files dictionary in the case
          that each column requires a different function mapping path
          to subject. If just one function is passed, it will be used
          for to load all dictionary entries. For example:

        | You may also pass the custom str 'auto' to
          specify that the subject name should be the base
          file name with the extension removed. For example
          if the path is '/some/path/subj16.npy' then the auto
          subject will be 'subj16'.

        | In the case that the underlying index is a MultiIndex, this
          function should be designed to return the subject in correct
          tuple form. See Examples below.

    {load_func}

    {inplace}

    See Also
    --------
    to_data_file : Cast existing columns to type Data File.
    get_file_mapping : Returns the raw file mapping.

    Examples
    ---------
    Consider the brief example below for loading two fake subjects,
    with the files parameter.

    ::

        files = dict()
        files['feat1'] = ['f1/subj_0.npy', 'f1/subj_1.npy']
        files['feat2'] = ['f2/subj_0.npy', 'f2/subj_1.npy']

    This could be matched with file_to_subject as:

    ::

        def file_to_subject_func(file):
            subject = file.split('/')[1].replace('.npy', '')
            return subject

        file_to_subject = file_to_subject_func
        # or
        file_to_subject = dict()
        file_to_subject['feat1'] = file_to_subject_func
        file_to_subject['feat2'] = file_to_subject_func

    In this example, subjects are loaded as 'subj_0' and 'subj_1',
    and they have associated loaded data files 'feat1' and 'feat2'.

    Next, we consider an example with fake data.
    In this example we will first generate and save some fake data files.
    These fake files will correspond to left hemisphere vertex files.

    .. ipython:: python

        import numpy as np
        import os

        dr = 'data/fake_surface/'
        os.makedirs(dr, exist_ok=True)

        # 20 subjects each with 10,242 vertex values
        X = np.random.random(size=(20, 10242))

        # Save the data as numpy arrays
        for x in range(len(X)):
            np.save(dr + str(x), X[x])

        os.listdir(dr)[:5]

    Next, we will use add data files to add these to
    a :class:`Dataset`.

    .. ipython:: python

        data = bp.Dataset()
        files = dict()
        files['fake_surface'] = dr + '*' # Add * for file globbing


        data = data.add_data_files(files=files, file_to_subject='auto')
        data.head(5)

    Let's also consider lastly a MultiIndex example:

    ::

        # The underlying dataset is indexed by subject and event
        data.set_index(['subject', 'event'], inplace=True)

        # Only one feature
        files = dict()
        files['feat1'] = ['f1/s0_e0.npy',
                          'f1/s0_e1.npy',
                          'f1/s1_e0.npy',
                          'f1/s1_e1.npy']

        def file_to_subject_func(file):

            # This selects the substring
            # at the last part seperated by the '/'
            # so e.g. the stub, 's0_e0.npy', 's0_e1.npy', etc...
            subj_split = file.split('/')[-1]

            # This removes the .npy from the end, so
            # stubs == 's0_e0', 's0_e1', etc...
            subj_split = subj_split.replace('.npy', '')

            # Set the subject name as the first part
            # and the eventname as the second part
            subj_name = subj_split.split('_')[0]
            event_name = subj_split.split('_')[1]

            # Lastly put it into the correct return style
            # This is tuple style e.g., ('s0', 'e0'), ('s0', 'e1')
            ind = (subj_name, eventname)

            return ind

    '''

    if not inplace:
        return self._inplace('add_data_files', locals())

    # Wrap load func if needed
    wrapped_load_func = wrap_load_func(load_func, _print=self._print)

    # Init if needed
    self._check_file_mapping()

    # Get dict of key to files
    file_series = proc_file_input(files, file_to_subject)

    # For each column
    for file in file_series:

        # For each subject, fill in with Data File
        series = file_series[file]
        self._series_to_data_file(col=file, series=series,
                                  load_func=wrapped_load_func)


@doc(**_sip_docs, load_func=_shared_docs['load_func'])
def to_data_file(self, scope,
                 load_func=np.load,
                 inplace=False):
    '''This method can be used to cast any existing columns
    where the values are file paths, to a data file.

    Parameters
    ----------
    {scope}

    {load_func}

    {inplace}

    Examples
    ----------
    This method can be used as a the primary way to prepare data files.
    We will perform a simple example here.

    .. ipython:: python

        import BPt as bp
        data = bp.Dataset()
        data['files'] = ['loc1.npy', 'loc2.npy']
        data

    We now have a :class:`Dataset`, but out column 'files' is not
    quite ready, as by default it won't know what to do with str.
    To get it to treat it as as a data file we will cast it.

    .. ipython:: python

        data = data.to_data_file('files')
        data

    What's happened here? Now it doesn't show paths anymore, but instead
    shows integers. That's actually the desired behavior though, we
    can check it out in file_mapping.

    .. ipython:: python

        data.file_mapping

    The file_mapping is then used internally with :class:`Loader`
    to load objects on the fly.

    See Also
    --------
    add_data_files : Method for adding new data files
    consolidate_data_files : Merge existing data files into one column.

    '''

    if not inplace:
        return self._inplace('to_data_file', locals())

    # Wrap load func if needed
    wrapped_load_func = wrap_load_func(load_func, _print=self._print)

    # Init if needed
    self._check_file_mapping()

    # Cast to data file
    for col in self.get_cols(scope):
        self._series_to_data_file(col=col, series=self[col],
                                  load_func=wrapped_load_func)


def _series_to_data_file(self, col, series, load_func):

    # Get next file mapping ind
    cnt = self._get_next_ind()

    for subject in series.index:

        # Create data file and add to file mapping
        data_file = DataFile(series[subject], load_func)
        self.file_mapping[cnt] = data_file

        # Replace cnt index in data
        self.at[subject, col] = cnt

        # Increment
        cnt += 1

    # Set scope
    self.add_scope(col, 'data file', inplace=True)


@doc(scope=_shared_docs['scope'])
def consolidate_data_files(self, save_dr, replace_with=None,
                           scope='data file', cast_to=None,
                           clear_existing='fail', n_jobs=-1):
    '''This function is designed as helper to consolidate all
    or a subset of the loaded data files into one column. While this
    removes information, in can provide a speed up in terms of downstream
    loading and reduce the number of files cached when using
    :class:`Loader`.

    This method assumes that the underlying data files
    can be stacked with ::

        np.stack(data, axis=-1)

    After they have been loaded. If this is not the case,
    then this function will break.

    Parameters
    -----------
    save_dr : str or Path
        The file directory in which to
        save the consolidated files. If it
        doesn't exist, then it will be created.

    replace_with : str or None, optional
        By default, if replace_with is left
        as None, then just a saved version of
        the files will be made. Instead,
        if a column name passed as a str is passed,
        then the original data files which were
        consolidated will be deleted, and the new
        consolidated column loaded instead.

        ::

            default = None

    {scope}
        ::

            default = 'data file'

    cast_to : None or numpy dtype, optional
        If not None, then this should be a
        numpy dtype in which the stacked data
        will be cast to before saving.

        ::

            default = None

    clear_existing : bool or 'fail', optional
        If True, then if the save dr already
        has files in it, delete them. If False,
        just overwrite them.

        If 'fail' then if there
        are already files in the save directory,
        raise an error.

        ::

            default = 'fail'

    n_jobs : int, optional
        The number of jobs to use while stacking
        and saving each file.

        If -1, then will try to use all avaliable cpu's.

        ::

            default == -1

    See Also
    --------
    to_data_file : Convert existing column to data file.
    add_data_files : Method for adding new data files
    update_data_file_paths : Update data path saved paths

    '''

    # Make sure file mapping up to date
    self._check_file_mapping()

    # If clear existing and exists
    if clear_existing is True:
        if os.path.exists(save_dr):
            self._print('Removing existing save directory:',
                        str(save_dr), level=0)
            shutil.rmtree(save_dr)

    # Make sure save_dr exists
    os.makedirs(save_dr, exist_ok=True)

    # If Fail.
    if clear_existing == 'fail':
        existing_files = len(os.listdir(save_dr))
        if existing_files > 0:
            raise RuntimeError('The save directory ' +
                               str(save_dr) + ' is not empty.'
                               ' Either change clear_existing or provide '
                               'a new save_dr.')

    # Get cols in scope
    cols = self.get_cols(scope)

    # For each subj / data point
    saved_locs = Parallel(n_jobs=n_jobs)(delayed(mp_consol_save)(
        data_files=[self.file_mapping[self.loc[index, key]]
                    for key in cols],
        index=index, cast_to=cast_to, save_dr=save_dr)
        for index in self.index)

    # If replace with
    if replace_with is not None:

        # Drop existing cols
        self.drop(cols, axis=1, inplace=True)

        # Create new series and add as new col
        self[replace_with] = pd.Series(saved_locs, index=self.index)

        # Cast to data file
        self.to_data_file(scope=replace_with,
                          inplace=True)


def update_data_file_paths(self, old, new):
    '''Go through and update saved file paths within
    the Datasets file mapping attribute.
    This function can be used
    when the underlying location of the data files has changed, or
    perhaps when loading a saved dataset on a different device.

    Note the old and new parameters work the same as those
    in the base python string.replace method.

    Parameters
    -----------
    old : str
        The substring in which to replace every instance found
        in every saved file path with new.

    new : str
        The substring in which to replace old with in
        every substring found.

    See Also
    --------
    to_data_file : Convert existing column to data file.
    add_data_files : Method for adding new data files
    '''

    self._check_file_mapping()

    for file_ind in self.file_mapping:
        self.file_mapping[file_ind].loc =\
            self.file_mapping[file_ind].loc.replace(old, new)


def _get_next_ind(self):

    if len(self.file_mapping) > 0:
        return np.nanmax(list(self.file_mapping.keys())) + 1
    else:
        return 0



def get_all_keys():

    # Start with a few used for testing errors and whatnot
    all_keys = set(['fake', 'default'])

    from .ensembles import all_obj_keys
    all_keys.update(all_obj_keys)

    from .feature_selectors import all_obj_keys
    all_keys.update(all_obj_keys)

    from .imputers import all_obj_keys
    all_keys.update(all_obj_keys)

    from .loaders import all_obj_keys
    all_keys.update(all_obj_keys)

    from .models import all_obj_keys
    all_keys.update(all_obj_keys)

    from .scalers import all_obj_keys
    all_keys.update(all_obj_keys)

    from .transformers import all_obj_keys
    all_keys.update(all_obj_keys)

    from ...extensions.mv_transformer.options import all_obj_keys
    all_keys.update(all_obj_keys)

    return all_keys



from .helpers import (update_mapping, proc_mapping, get_reverse_mapping)
import numpy as np
from joblib import Parallel, delayed
from sklearn.utils.validation import check_memory
from sklearn.base import clone
from .ScopeObjs import ScopeTransformer
from operator import itemgetter
from .base import _get_est_trans_params
from ..util import get_top_substrs
import pandas as pd


def load_and_trans(transformer, load_func, loc):
    '''This function is designed to be able to be wrapped in a cache
    check_memory.'''

    data = load_func(loc)
    trans_data = np.squeeze(transformer.fit_transform(data))
    return trans_data


def get_trans_chunk(transformer, data_files, func):
    '''This function is designed to be used for multi-processing'''

    X_trans_chunk = []
    for DataFile in data_files:
        loc = DataFile.loc
        load_func = DataFile.load_func
        trans_data = func(clone(transformer), load_func, loc)
        X_trans_chunk.append(trans_data)

    return X_trans_chunk


class BPtLoader(ScopeTransformer):

    # Override
    _required_parameters = ["estimator", "inds", "file_mapping"]

    def __init__(self, estimator, inds, file_mapping,
                 n_jobs=1, fix_n_jobs=False,
                 cache_loc=None):
        '''The inds for loaders are special, they should not be
        set with Ellipsis. Instead in the case of all, should be
        passed inds as usual.'''

        # Set Super params
        super().__init__(estimator=estimator, inds=inds, cache_loc=cache_loc)

        # Set rest of params
        self.file_mapping = file_mapping

        # Make sure to set fix n jobs before n_jobs
        self.fix_n_jobs = fix_n_jobs
        self.n_jobs = n_jobs

    # Override inherited n_jobs propegate behavior
    @property
    def n_jobs(self):
        return self.n_jobs_proxy

    @n_jobs.setter
    def n_jobs(self, n_jobs):
        self.n_jobs_proxy = n_jobs

    @property
    def _n_jobs(self):

        if self.fix_n_jobs is False:
            return self.n_jobs

        return self.fix_n_jobs

    def fit(self, X, y=None, mapping=None,
            fit_index=None, **fit_params):

        # Need the output from a transform to full fit,
        # so when fit is called, call fit_transform instead
        self.fit_transform(X=X, y=y, mapping=mapping,
                           fit_index=fit_index,
                           **fit_params)

        return self

    def _fit(self, X, y=None, **fit_params):
        '''Special fit for loader, only concerned with
        fitting first data point for e.g., use of reverse
        transform
        '''

        # Get the first data point
        first_feat = self.inds_[0]
        fit_fm_key = X[0, first_feat]
        fit_X = self.file_mapping[int(fit_fm_key)].load()

        # Fit the first data point
        self.estimator_.fit(fit_X, y=y, **fit_params)

    def _update_loader_mappings(self, mapping):

        # Need to set out_mapping_ and update mapping
        self.out_mapping_ = {}

        # Add changed X_trans by col
        for c in range(len(self.inds_)):
            ind = self.inds_[c]
            self.out_mapping_[ind] = self.X_trans_inds_[c]

        # Fill the remaining spots sequentially,
        # for each of the rest inds.
        for c in range(len(self.rest_inds_)):
            ind = self.rest_inds_[c]
            self.out_mapping_[ind] = self.n_trans_feats_ + c

        # Update the original mapping, this is the mapping which
        # will be passed to the next piece of the pipeline
        update_mapping(mapping, self.out_mapping_)

    def fit_transform(self, X, y=None, mapping=None,
                      fit_index=None, **fit_params):

        if mapping is None:
            mapping = {}

        # Call parent fit but passing only the first data point
        super().fit(X, y=y, mapping=mapping,
                    fit_index=fit_index,
                    **fit_params)

        # If skipped, skip
        if self.estimator_ is None:
            return X

        # The parent fit takes care of, in addition to
        # fitting the loader on one
        # data point, sets base_dtype, processes the mapping,
        # sets rest inds, etc...

        # Now transform X - this sets self.X_trans_inds_
        X_trans = self.transform(X, transform_index=fit_index)

        # Update the mapping + out_mapping_
        self._update_loader_mappings(mapping)

        # Now return X_trans
        return X_trans

    def transform(self, X, transform_index=None):

        # Skip if skipped
        if self.estimator_ is None:
            return X

        # @ TODO transform index just exists for compat
        # with loader right now, won't actually propegate.

        # Init lists + mappings
        X_trans, self.X_trans_inds_ = [], []
        cnt = 0

        # For each column to fit_transform
        for col in self.inds_:

            # Get transformer column
            fm_keys = [key for key in X[:, col]]
            X_trans_cols = self._get_trans_col(fm_keys)

            # Stack + append new features
            X_trans_cols = np.stack(X_trans_cols)
            X_trans.append(X_trans_cols)

            # Add + append inds
            X_trans_cols_inds =\
                [i for i in range(cnt, X_trans_cols.shape[1] + cnt)]
            self.X_trans_inds_.append(X_trans_cols_inds)

            # Update cnt
            cnt = X_trans_cols.shape[1] + cnt

        # Stack final
        X_trans = np.hstack(X_trans)

        # Save number of output features after X_trans
        self.n_trans_feats_ = X_trans.shape[1]

        # Return stacked X_trans with rest inds
        return np.hstack([X_trans, X[:, self.rest_inds_]])

    def get_chunks(self, data_files):

        per_chunk = len(data_files) // self._n_jobs
        chunks = [list(range(i * per_chunk, (i+1) * per_chunk))
                  for i in range(self._n_jobs)]

        last = chunks[-1][-1]
        chunks[-1] += list(range(last+1, len(data_files)))
        return [[data_files[i] for i in c] for c in chunks]

    def _get_trans_col(self, fm_keys):

        # Grab the right data files from the file mapping (casting to int!)
        data_files = [self.file_mapping[int(fm_key)] for fm_key in fm_keys]

        # Clone the base loader
        cloned_estimator = clone(self.estimator)

        # If a caching location is passed, create new load_and_trans_c func
        if self.cache_loc is not None:
            memory = check_memory(self.cache_loc)
            load_and_trans_c = memory.cache(load_and_trans)
        else:
            load_and_trans_c = load_and_trans

        if self._n_jobs == 1:
            X_trans_cols = get_trans_chunk(cloned_estimator,
                                           data_files, load_and_trans_c)
        else:
            chunks = self.get_chunks(data_files)

            X_trans_chunks =\
                Parallel(n_jobs=self._n_jobs)(
                    delayed(get_trans_chunk)(
                        transformer=cloned_estimator,
                        data_files=chunk,
                        func=load_and_trans_c)
                    for chunk in chunks)

            X_trans_cols = []
            for chunk in X_trans_chunks:
                X_trans_cols += chunk

        return X_trans_cols

    def transform_df(self, df, base_name='loader', encoders=None):

        return super().transform_df(df, base_name=base_name)

    def _proc_new_names(self, feat_names, base_name=None, encoders=None):

        # If skip, return passed names as is
        if self.estimator_ is None:
            return feat_names

        # Store original passed feat names here
        self.feat_names_in_ = feat_names

        # If skip, return passed names as is
        if self.estimator_ is None:
            return feat_names

        # If base loader has stored feat names, use those.
        if hasattr(self.estimator_, 'feat_names_'):
            return getattr(self.estimator_, 'feat_names_')

        # Get new names
        new_names = []
        for c in range(len(self.inds_)):

            ind = self.inds_[c]
            col_name = feat_names[ind]

            new_inds = self.X_trans_inds_[c]
            new_names += [str(col_name) + '_' + str(i)
                          for i in range(len(new_inds))]

        # Remove old names - using parent method
        feat_names = self._remove_old_names(feat_names)

        # New names come first, then rest of names
        all_names = new_names + feat_names

        return all_names

    def inverse_transform_FIs(self, fis):

        # Skip if skipped
        if self.estimator_ is None:
            return fis

        # If doesn't have inverse_transform, return as is.
        if not hasattr(self.estimator_, 'inverse_transform'):
            return fis

        # Get feature importances also as array
        fis_data = np.array(fis)
        fis_names = np.array(fis.index)

        # Compute reverse mapping
        reverse_mapping = get_reverse_mapping(self.mapping_)

        # Prep return fis
        return_fis_data = np.zeros(len(reverse_mapping), dtype='object')
        return_fis_names = ['' for _ in range(len(reverse_mapping))]

        # Process each feature
        for col_ind in self.inds_:

            # Get reverse inds and data for just this col
            reverse_inds = proc_mapping([col_ind], self.out_mapping_)
            col_fis = fis_data[reverse_inds]
            col_names = fis_names[reverse_inds]

            # Run inverse transform
            inv_fis = self.estimator_.inverse_transform(col_fis)

            # Place into return fis_data
            original_ind = reverse_mapping[col_ind]
            return_fis_data[original_ind] = inv_fis

            # Add return name
            return_fis_names[original_ind] =\
                self._get_reverse_feat_name(col_names, original_ind)

        # Fill in with original
        for col_ind in self.rest_inds_:
            reverse_ind = proc_mapping([col_ind], self.out_mapping_)[0]
            original_ind = reverse_mapping[col_ind]

            # Just pass along data and name, but in original spot
            return_fis_data[original_ind] = fis_data[reverse_ind]
            return_fis_names[original_ind] = fis_names[reverse_ind]

        # Return new series
        return pd.Series(return_fis_data, index=return_fis_names)

    def _get_reverse_feat_name(self, col_names, original_ind):

        # Get + add return name
        if hasattr(self, 'feat_names_in_'):
            return self.feat_names_in_[original_ind]

        # If reverse_feat_names hasn't been called
        substrs = get_top_substrs(col_names)

        # If no common sub string
        if len(substrs) == 0:
            return 'loader_ind_' + str(original_ind)

        # Just take first substring
        new_name = substrs[0]

        # Remove ending _ if any
        if new_name.endswith('_'):
            new_name = new_name[:-1]

        return new_name


class CompatArray(list):

    def __init__(self, arr_2d):

        self.dtype = arr_2d.dtype
        super().__init__(np.swapaxes(arr_2d, 0, 1))

    @property
    def shape(self):
        return (len(self[0]), len(self))

    def conv_rest_back(self, rest_inds):

        if len(rest_inds) == 0:
            empty = np.array([], dtype=self.dtype)
            return empty.reshape((self.shape[0], 0))

        # Create an array from the requested rest inds
        base = np.array(itemgetter(*rest_inds)(self),
                        dtype=self.dtype)

        # If only one return axis, conv to correct shape
        if len(base.shape) == 1:
            return base[:, np.newaxis]

        # Reverse initial swap
        return np.swapaxes(base, 0, 1)


class BPtListLoader(BPtLoader):

    def fit_transform(self, X, y=None, mapping=None,
                      fit_index=None, **fit_params):

        # Process the mapping
        if mapping is None:
            mapping = {}
        self._proc_mapping(mapping)

        if len(self.inds_) != 1:
            raise RuntimeWarning('BPtListLoader can only work on one column.')

        # Calls super fit_transform, but passing
        # X with the data columns replaced by CompatArray
        return super().fit_transform(self._get_X_compat(X), y=y,
                                     mapping=mapping,
                                     fit_index=fit_index,
                                     **fit_params)

    def _get_X_compat(self, X):

        ind = self.inds[0]
        X_loaded = CompatArray(X)
        X_loaded[ind] = self._load_col(X[:, ind])

        return X_loaded

    def _load_col(self, X_col):
        '''Load X col as list of subjects'''

        X_col_loaded = []
        for key in X_col:
            DataFile = self.file_mapping[int(key)]
            data = DataFile.load()
            X_col_loaded.append(data)

        return X_col_loaded

    def _fit(self, X, y=None, **fit_params):
        '''Override the internal fit function to fit only
        the single requested column.'''
        self.estimator_.fit(X[self.inds_[0]], y=y, **fit_params)

        return self

    def transform(self, X, transform_index=None):

        # If None, pass along as is
        if self.estimator_ is None:
            return X

        # Load if not loaded
        if not isinstance(X, CompatArray):
            X = self._get_X_compat(X)

        # Get transform params
        trans_params = _get_est_trans_params(self.estimator_,
                                             transform_index=transform_index)

        # Get X_trans
        X_trans = self.estimator_.transform(X=X[self.inds_[0]], **trans_params)

        # Save number of output features after X_trans
        self.n_trans_feats_ = X_trans.shape[1]

        # For compat
        self.X_trans_inds_ = [list(range(self.n_trans_feats_))]

        # Return stacked X_trans with rest inds
        return np.hstack([X_trans, X.conv_rest_back(self.rest_inds_)])


from sklearn.base import BaseEstimator
from .helpers import get_possible_params


def _needs(estimator, flag, attr, method):

    # If has class attribute flag
    if hasattr(estimator, flag):
        if getattr(estimator, flag):
            return True

    # Next check if matches name of method param
    elif attr in get_possible_params(estimator, method):
        return True

    # Otherwise False
    return False


def _get_est_fit_params(estimator, mapping=None,
                        fit_index=None,
                        other_params=None,
                        copy_mapping=True):

    if other_params is None:
        fit_params = {}
    else:
        fit_params = other_params.copy()

    if 'mapping' in fit_params and mapping is None:
        mapping = fit_params.pop('mapping')

    if 'fit_index' in fit_params and fit_index is None:
        fit_index = fit_params.pop('fit_index')

    if mapping is not None and _needs(estimator, '_needs_mapping',
                                      'mapping', 'fit'):

        if copy_mapping:
            fit_params['mapping'] = mapping.copy()
        else:
            fit_params['mapping'] = mapping

    if fit_index is not None and _needs(estimator, '_needs_fit_index',
                                        'fit_index', 'fit'):
        fit_params['fit_index'] = fit_index

    return fit_params


def _get_est_trans_params(estimator, transform_index=None):

    trans_params = {}

    if transform_index is not None and _needs(estimator,
                                              '_needs_transform_index',
                                              'transform_index', 'transform'):
        trans_params['transform_index'] = transform_index

    return trans_params


def _get_est_fit_trans_params(estimator, mapping=None,
                              fit_index=None,
                              other_params=None,
                              copy_mapping=True):

    # Get fit params first
    fit_trans_params =\
        _get_est_fit_params(estimator,
                            mapping=mapping,
                            fit_index=fit_index,
                            other_params=other_params,
                            copy_mapping=copy_mapping)

    # Then update transform params,
    # note that since it is fit_transform, the transform index
    # is the same as the fit_index here
    fit_trans_params.update(
        _get_est_trans_params(estimator,
                              transform_index=fit_index))

    return fit_trans_params


def _fit_single_estimator(estimator, X, y, sample_weight=None,
                          mapping=None, fit_index=None,
                          message_clsname=None, message=None):
    """Private function used to fit an estimator within a job."""

    fit_params = _get_est_fit_params(estimator=estimator, mapping=mapping,
                                     fit_index=fit_index)

    if sample_weight is not None:
        try:
            estimator.fit(X, y, sample_weight=sample_weight, **fit_params)
        except TypeError as exc:
            if "unexpected keyword argument 'sample_weight'" in str(exc):
                raise TypeError(
                    "Underlying estimator {} does not support sample weights."
                    .format(estimator.__class__.__name__)
                ) from exc
            raise
    else:
        estimator.fit(X, y, **fit_params)

    return estimator


class BPtBase(BaseEstimator):

    _required_parameters = ["estimator"]

    def __init__(self, estimator):
        self.estimator = estimator

    @property
    def n_jobs(self):
        if hasattr(self.estimator, 'n_jobs'):
            return getattr(self.estimator, 'n_jobs')

    @n_jobs.setter
    def n_jobs(self, n_jobs):
        if hasattr(self.estimator, 'n_jobs'):
            setattr(self.estimator, 'n_jobs', n_jobs)

    def transform_df(self, X_df, base_name=None, encoders=None):
        '''Override this class in child classes'''
        pass



from ..params.Params import TransitionChoice, Choice, Dict, Scalar, Array
import nevergrad as ng


def test_content():

    base = ng.p.Choice([None, 'balanced'])
    p = Choice([None, 'balanced'])
    assert repr(base._content) == repr(p._content)


def test_transition_choice():

    cls_weight = TransitionChoice([None, 'balanced'],
                                  transitions=[1, 1])

    assert repr(cls_weight) == \
        "TransitionChoice([None, 'balanced'], transitions=[1, 1])"

    # Test nested case
    g = TransitionChoice([cls_weight, Dict(q=cls_weight)])
    assert hasattr(g.choices[0], 'name')
    assert hasattr(g.choices[1], 'name')
    assert hasattr(g.choices[1]['q'], 'name')

    r = "TransitionChoice([TransitionChoice([None, 'balanced'], " + \
        "transitions=[1, 1]), Dict(q=TransitionChoice([None, 'balanced']," + \
        " transitions=[1, 1]))])"
    assert repr(g) == r


def test_scalar():

    # Confirm base behavior
    base = ng.p.Scalar(lower=5, upper=10).set_integer_casting()
    assert repr(base).startswith('Scalar{int')
    assert base.integer

    # Confirm cast works
    p = Scalar(lower=5, upper=10).set_integer_casting()
    assert p.integer


def test_array():

    base = ng.p.Array(init=(100, 100)).set_mutation(sigma=50)
    assert base.sigma._value == 50

    p = Array(init=(100, 100)).set_mutation(sigma=50)
    assert p.sigma._value == 50

    base.set_bounds(lower=1, upper=300)
    assert base.bounds[0][0] == 1
    assert base.bounds[1][0] == 300

    p = Array(init=(100, 100)).set_mutation(sigma=50)
    p.set_bounds(lower=1, upper=300)

    assert p.bounds[0][0] == 1
    assert p.bounds[1][0] == 300

    repr_p = "Array(init=(100, 100)).set_mutation(sigma=50)." + \
        "set_bounds(lower=1, upper=300)"

    assert repr(p) == repr_p


def test_to_grid():

    params = {'1': Choice([1, 2, 3]),
              '2': TransitionChoice([1, 2]),
              '3': 3,
              '4': Scalar(lower=1, upper=2).set_integer_casting()}

    assert params['1'].to_grid() == [1, 2, 3]
    assert params['2'].to_grid() == [1, 2]
    assert params['4'].to_grid() == [1, 2]

    assert hasattr(params['1'], 'to_grid')
    assert hasattr(params['2'], 'to_grid')



from sklearn.feature_selection._base import SelectorMixin
from .helpers import update_mapping, get_reverse_mapping, proc_mapping
from .ScopeObjs import ScopeTransformer
import numpy as np
import pandas as pd


class BPtFeatureSelector(ScopeTransformer, SelectorMixin):

    def _update_feat_mapping(self, X, mapping):

        # Need to pass along the correct mapping
        # overwrite existing out mapping
        self.out_mapping_ = {}

        # This is the calculated support from the base estimator
        support = self.estimator_.get_support()

        # Set in scope inds by if all case
        if self.inds_ is Ellipsis:
            in_scope_inds = list(range(X.shape[1]))
        else:
            in_scope_inds = self.inds_

        # First half is for updating the index within scope
        cnt = 0
        for i, ind in enumerate(in_scope_inds):

            # If kept by feat selection, add, otherwise set to None
            if support[i]:
                self.out_mapping_[ind] = cnt
                cnt += 1
            else:
                self.out_mapping_[ind] = None

        # Next, need to update the mapping for the remaining wrapper inds
        # essentially setting them where the cnt left off, then sequentially
        # If None, will just  skip
        for rem_ind in range(len(self.rest_inds_)):
            self.out_mapping_[self.rest_inds_[rem_ind]] = cnt
            cnt += 1

        # Update the original mapping, this is the mapping which
        # will be passed to the next piece of the pipeline
        update_mapping(mapping, self.out_mapping_)

        return self

    def fit(self, X, y=None, mapping=None,
            fit_index=None, **fit_params):

        if mapping is None:
            mapping = {}

        # Call parent fit
        super().fit(X, y=y, mapping=mapping,
                    fit_index=fit_index,
                    **fit_params)

        # Need to update mapping
        return self._update_feat_mapping(X, mapping)

    def _proc_new_names(self, feat_names, base_name=None, encoders=None):

        # If skip, return passed names as is
        if self.estimator_ is None:
            return feat_names

        # Store original passed feat names here
        self.feat_names_in_ = feat_names

        # Get base new names from parent class
        new_names = super()._proc_new_names(feat_names)

        # This feat mask corresponds to the already transformed feats
        feat_mask = self._get_support_mask()

        # Apply the computed mask to get the actually selected features
        return_names = np.array(new_names)[feat_mask]

        return list(return_names)

    def _get_support_mask(self):

        # Create full support as base support + True's for all rest inds
        # i.e., those features originally out of scope
        base_support = self.estimator_.get_support()
        rest_support = np.ones(len(self.rest_inds_), dtype='bool')
        support = np.concatenate([base_support, rest_support])

        return support

    def inverse_transform_FIs(self, fis):

        # Skip if skipped
        if self.estimator_ is None:
            return fis

        # Get data as input form
        fis_data = np.array(fis).reshape(1, -1)

        # Get return data from inverse transform
        return_fis_data = self.estimator_.inverse_transform(fis_data)[0]

        if not hasattr(self, 'feat_names_in_'):
            raise RuntimeError('_proc_new_names must be called first.')

        # Put in a series to return
        return_fis = pd.Series(return_fis_data,
                               index=self.feat_names_in_)

        return return_fis



"""
Feature_Selectors.py
====================================
File with different Feature Selectors
"""
from ..helpers import (get_possible_init_params,
                       get_obj_and_params, all_from_avaliable)
from sklearn.feature_selection import SelectPercentile, VarianceThreshold
from ...extensions.FeatSelectors import FeatureSelector
from sklearn.feature_selection import RFE
import numpy as np
from numpy.random import RandomState
from ..params.Params import Array


AVALIABLE = {
        'binary': {
                'univariate selection c':
                'univariate selection c',
                'univariate selection':
                'univariate selection c',
                'rfe': 'rfe',
                'variance threshold': 'variance threshold',
                'selector': 'selector',
        },
        'regression': {
                'univariate selection r':
                'univariate selection r',
                'univariate selection':
                'univariate selection r',
                'rfe': 'rfe',
                'variance threshold': 'variance threshold',
                'selector': 'selector',
        },
}

AVALIABLE['categorical'] = AVALIABLE['binary'].copy()

SELECTORS = {
    'univariate selection r': (SelectPercentile,
                               ['base univar fs regression',
                                'univar fs regression dist',
                                'univar fs r keep more',
                                'univar fs r keep less']),

    'univariate selection c': (SelectPercentile,
                               ['base univar fs classifier',
                                'univar fs classifier dist',
                                'univar fs c keep more',
                                'univar fs c keep less']),

    'rfe': (RFE, ['base rfe', 'rfe num feats dist']),

    'variance threshold': (VarianceThreshold, ['default']),

    'selector': (FeatureSelector, ['random', 'searchable']),
}


def get_special_selector(feat_selector, feat_selector_params,
                         random_state, num_feat_keys):

    # Init feat selector with mask of random feats
    if random_state is None:
        r_state = RandomState(np.random.randint(1000))
    elif isinstance(random_state, int):
        r_state = RandomState(random_state)
    else:
        r_state = random_state

    init_mask = r_state.random(num_feat_keys)
    feat_selector = feat_selector(mask=init_mask)

    # Figure out param passed
    if 'selector__mask' in feat_selector_params:
        p_name = 'selector__mask'

        # If set to searchable, set to searchable...
        if feat_selector_params[p_name] == 'sets as hyperparameters':

            feat_array = Array(init=[.5 for i in range(num_feat_keys)])
            feat_array.set_mutation(sigma=1/6).set_bounds(lower=0, upper=1)
            feat_selector_params[p_name] = feat_array

        elif feat_selector_params[p_name] == 'sets as random features':
            del feat_selector_params[p_name]

    return feat_selector, feat_selector_params


def get_feat_selector_and_params(feat_selector_str, extra_params, params,
                                 random_state, num_feat_keys):
    '''Returns a scaler based on proc'ed str indicator input,

    Parameters
    ----------
    feat_selector_str : str
        `feat_selector_str` refers to the type of feature selection
        to apply to the saved data during model evaluation.

    extra_params : dict
        Any extra params being passed.
        These can be supplied by creating another dict within extra_params.
        E.g., extra_params[method] = {'method param' : new_value}
        Where method param is a valid argument for that method,
        and method in this case is the str indicator.

    params : int
        The index of the params to use.

    Returns
    ----------
    feature_selector
        A feature selector object with fit and transform methods.

    dict
        The params for this feature selector
    '''

    # Base behavior
    feat_selector, extra_feat_selector_params, feat_selector_params =\
        get_obj_and_params(feat_selector_str, SELECTORS, extra_params, params)

    # Special behavior for selector...
    if feat_selector_str == 'selector':

        # Move to params
        if 'mask' in extra_feat_selector_params:
            feat_selector_params['selector__mask'] =\
                extra_feat_selector_params.pop('mask')

        feat_selector, feat_selector_params =\
            get_special_selector(feat_selector, feat_selector_params,
                                 random_state, num_feat_keys)

        return feat_selector, feat_selector_params

    # Otherwise, need to check for estimator
    # as RFE needs a default param for est.
    # Set as placeholder None if passed
    possible_params = get_possible_init_params(feat_selector)
    if 'estimator' in possible_params:
        extra_feat_selector_params['estimator'] = None

    return (feat_selector(**extra_feat_selector_params),
            feat_selector_params)


all_obj_keys = all_from_avaliable(AVALIABLE)



from ..Dataset import Dataset
import numpy as np


def get_fake_dataset():
    '''
    fake['1'] = [1, 2, 3]
    fake['2'] = ['6', '7', '8']
    fake['2'] = fake['2'].astype('category')
    fake['3'] = [np.nan, 2, 3]
    '''

    fake = Dataset()
    fake.verbose = -1
    fake['1'] = [1, 2, 3]
    fake['2'] = ['6', '7', '8']
    fake['2'] = fake['2'].astype('category')
    fake['3'] = [np.nan, 2, 3]

    return fake


def get_fake_dataset2():

    fake = Dataset()
    fake['1'] = [1, 1, 1]
    fake['2'] = [1, 2, 3]
    fake['3'] = ['1', '2', '3']

    return fake


def get_fake_dataset3():

    fake = Dataset()
    fake['1'] = [1, 1, 1]
    fake['2'] = [1, 1, 1]
    fake['3'] = ['2', '2', '2']
    fake['4'] = ['2', '2', '2']
    fake['5'] = ['2', 1, '2']

    return fake


def get_fake_dataset4():

    fake = Dataset()
    fake['1'] = [1, 2, 3, 4, 5, 6]
    fake['2'] = [1, 2, 3, 4, 5, np.nan]

    return fake


def get_fake_dataset5():

    df = Dataset()
    df['1'] = [1, 2, 3, 4, 5]
    df['2'] = [1, 1, 1, 1, 2]
    df['3'] = [np.nan, np.nan, 1, 1, 2]
    df['4'] = [1, 2, 2, 3, 3]
    df['4'] = df['4'].astype('object')
    return df


def get_fake_dataset6():

    df = Dataset()
    df['1'] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 1]
    df['2'] = [0, 0, 0, 0, 0, 0, 0, 0, np.nan, 1]
    return df


def get_fake_dataset7():

    df = Dataset()
    df['1'] = [0, 1, 1, 1, 2, 2, 2]
    df['1'] = df['1'].astype('category')

    df['2'] = [0, 1, 1, 1, 2, 2, 2]

    return df


def get_nans_dataset():

    fake = Dataset()
    fake['1'] = [np.nan, np.nan, np.nan, 1]
    fake['2'] = [np.nan, np.nan, 1, 1]
    fake['3'] = [np.nan, 1, 1, 1]
    fake['4'] = [1, 1, 1, 1]
    return fake


def get_full_dataset():

    fake = Dataset()
    fake['1'] = [1, 2, 3, 4, 5]
    fake['2'] = [6, 7, 8, 9, 10]
    fake['3'] = [11, 12, 13, 14, 15]
    fake.add_scope('3', 'category', inplace=True)

    fake['subj'] = ['s1', 's2', 's3', 's4', 's5']
    fake.set_index('subj', inplace=True)

    fake['targ'] = [.1, .2, .3, .4, .5]
    fake = fake.set_role('targ', 'target')

    fake = fake.set_test_split(subjects=['s4', 's5'])

    return fake


def get_full_int_index_dataset():

    fake = Dataset()
    fake['1'] = [1, 2, 3, 4, 5]
    fake['2'] = [6, 7, 8, 9, 10]
    fake['3'] = [11, 12, 13, 14, 15]
    fake.add_scope('3', 'category', inplace=True)

    fake['subj'] = [1, 2, 3, 4, 5]
    fake.set_index('subj', inplace=True)

    fake['targ'] = [.1, .2, .3, .4, .5]
    fake = fake.set_role('targ', 'target')

    fake = fake.set_test_split(subjects=[4, 5])

    return fake


def get_fake_multi_index_dataset():

    fake = Dataset()
    fake['0'] = [1, 2, 3, 4, 5, 6]
    fake['1'] = [1, 2, 3, 4, 5, 6]
    fake['2'] = [1, 2, 3, 4, 5, np.nan]
    fake['subj'] = ['s1', 's2', 's3', 's1', 's2', 's3']
    fake['event'] = ['e1', 'e1', 'e1', 'e2', 'e2', 'e2']
    fake.set_index(['subj', 'event'], inplace=True)

    return fake



import numpy as np
from ..Dataset import Dataset
import pytest
import os
import tempfile
from ...main.input import CVStrategy


def get_fake_dataset():
    '''
    fake = Dataset()
    fake['1'] = [1, 1, 1]
    fake['2'] = [1, 1, 1]
    fake['3'] = ['2', '2', '2']
    fake['4'] = ['2', '2', '2']
    fake['5'] = ['2', '1', '2']

    fake = fake.set_role('1', 'data')
    fake = fake.set_role('2', 'target')
    fake = fake.set_role(['3', '4', '5'], 'non input')

    fake.ordinalize(scope='all', inplace=True)
    '''

    fake = Dataset()
    fake['1'] = [1, 1, 1]
    fake['2'] = [1, 1, 1]
    fake['3'] = ['2', '2', '2']
    fake['4'] = ['2', '2', '2']
    fake['5'] = ['2', '1', '2']

    fake = fake.set_role('1', 'data')
    fake = fake.set_role('2', 'target')
    fake = fake.set_role(['3', '4', '5'], 'non input')

    fake.ordinalize(scope='all', inplace=True)

    return fake


def test_proc_cv_strategy_base():

    df = get_fake_dataset()

    cv_params = CVStrategy()

    cv = df._proc_cv_strategy(cv_params)
    assert cv.groups is None
    assert cv.stratify is None
    assert cv.train_only is None

    cv = df._proc_cv_strategy(cv_params=None)
    assert cv.groups is None
    assert cv.stratify is None
    assert cv.train_only is None


def test_proc_cv_strategy_train_only():

    df = get_fake_dataset()

    cv_params = CVStrategy(train_only_subjects=[0, 1])
    cv = df._proc_cv_strategy(cv_params)

    assert cv.groups is None
    assert cv.stratify is None
    assert np.array_equal(cv.train_only, np.array([0, 1]))

    # Make sure sorts for repeatable behavior
    cv_params = CVStrategy(train_only_subjects=[0, 1])
    cv = df._proc_cv_strategy(cv_params)

    assert cv.groups is None
    assert cv.stratify is None
    assert np.array_equal(cv.train_only, np.array([0, 1]))


def test_proc_cv_strategy_groups():

    df = get_fake_dataset()

    with pytest.raises(RuntimeError):
        cv_params = CVStrategy(groups=['1', '2'])

    cv_params = CVStrategy(groups='1')
    with pytest.raises(RuntimeError):
        cv = df._proc_cv_strategy(cv_params)

    cv_params = CVStrategy(groups="doesn't exist")
    with pytest.raises(KeyError):
        cv = df._proc_cv_strategy(cv_params)

    cv_params = CVStrategy(groups='2')
    with pytest.raises(RuntimeError):
        cv = df._proc_cv_strategy(cv_params)

    cv_params = CVStrategy(groups='3')
    cv = df._proc_cv_strategy(cv_params)
    assert len(cv.groups) == 3
    assert cv.groups.nunique() == 1


def test_proc_cv_strategy_stratify():

    df = get_fake_dataset()

    with pytest.raises(RuntimeError):
        cv_params = CVStrategy(stratify=['1', '2'])

    cv_params = CVStrategy(stratify='1')
    with pytest.raises(RuntimeError):
        cv = df._proc_cv_strategy(cv_params)

    cv_params = CVStrategy(stratify="doesn't exist")
    with pytest.raises(KeyError):
        cv = df._proc_cv_strategy(cv_params)

    cv_params = CVStrategy(stratify='2')
    cv = df._proc_cv_strategy(cv_params)
    assert len(cv.stratify) == 3
    assert cv.stratify.nunique() == 1

    cv_params = CVStrategy(stratify='3')
    cv = df._proc_cv_strategy(cv_params)
    assert len(cv.stratify) == 3
    assert cv.stratify.nunique() == 1


def test_set_test_split():

    df = get_fake_dataset()
    df.verbose = -1

    with pytest.raises(TypeError):
        df = df.set_test_split()

    with pytest.raises(TypeError):
        df = df.set_test_split(size=.2, subjects=[1, 2])

    df = df.set_test_split(size=1, cv_strategy=None, random_state=None)

    assert len(df.test_subjects) == 1
    assert len(df.train_subjects) == 2

    df = df.set_test_split(size=.3, cv_strategy=None, random_state=None)

    assert len(df.test_subjects) == 1
    assert len(df.train_subjects) == 2

    df.set_test_split(size=.5,
                      cv_strategy=CVStrategy(train_only_subjects=[0]),
                      random_state=1, inplace=True)

    assert len(df.test_subjects) == 1
    assert len(df.train_subjects) == 2
    assert 0 in df.train_subjects

    df.set_test_split(size=1, cv_strategy=CVStrategy(train_only_subjects=[0]),
                      random_state=1, inplace=True)

    assert len(df.test_subjects) == 1
    assert len(df.train_subjects) == 2
    assert 0 in df.train_subjects

    temp_loc = os.path.join(tempfile.gettempdir(), 'temp.txt')
    df.save_test_split(temp_loc)
    df = df.set_test_split(subjects=temp_loc)
    os.remove(temp_loc)

    assert len(df.test_subjects) == 1
    assert len(df.train_subjects) == 2
    assert 0 in df.train_subjects


def test_set_train_split():

    df = get_fake_dataset()

    with pytest.raises(TypeError):
        df = df.set_train_split()

    with pytest.raises(TypeError):
        df = df.set_train_split(size=.2, subjects=[1, 2])

    df = df.set_train_split(size=1, cv_strategy=None, random_state=None)
    assert len(df.test_subjects) == 2
    assert len(df.train_subjects) == 1

    df = df.set_train_split(size=.4, cv_strategy=None, random_state=None)
    assert len(df.test_subjects) == 2
    assert len(df.train_subjects) == 1

    df.set_train_split(size=.5,
                       cv_strategy=CVStrategy(train_only_subjects=[0]),
                       random_state=1, inplace=True)
    assert len(df.test_subjects) == 1
    assert len(df.train_subjects) == 2
    assert 0 in df.train_subjects

    df.set_train_split(size=1,
                       cv_strategy=CVStrategy(train_only_subjects=[0]),
                       random_state=1, inplace=True)
    assert len(df.test_subjects) == 1
    assert len(df.train_subjects) == 2
    assert 0 in df.train_subjects

    # Test save and load
    temp_loc = os.path.join(tempfile.gettempdir(), 'temp.txt')
    df.save_train_split(temp_loc)
    df = df.set_train_split(subjects=temp_loc)
    os.remove(temp_loc)

    assert len(df.test_subjects) == 1
    assert len(df.train_subjects) == 2
    assert 0 in df.train_subjects

    with pytest.raises(ValueError):
        df.set_train_split(size=1,
                           cv_strategy=CVStrategy(train_only_subjects=[0, 1]),
                           random_state=1, inplace=True)

    with pytest.raises(RuntimeError):
        df = df.set_train_split(size=0)


def test_test_split():

    df = get_fake_dataset()
    df.add_scope('1', 'test_scope', inplace=True)

    tr_df, test_df = df.test_split(subjects=[2])

    # Check copy data behavior
    original = tr_df.loc[0, '5']
    tr_df.loc[0, '5'] = 0
    assert tr_df.loc[0, '5'] != df.loc[0, '5']
    assert original == df.loc[0, '5']

    # Check some assumptions
    assert tr_df.shape == (2, 5)
    assert 0 in tr_df.index
    assert 1 in tr_df.index

    assert test_df.shape == (1, 5)
    assert 2 in test_df.index

    for d in [tr_df, test_df]:
        assert isinstance(d, Dataset)
        assert d.test_subjects is None
        assert d.train_subjects is None

        scopes = d.get_scopes()
        assert 'test_scope' in scopes['1']

    # Check meta data behavior
    tr_df.add_scope('1', 'train', inplace=True)
    scopes = df.get_scopes()
    assert 'train' not in scopes['1']

    test_scopes = test_df.get_scopes()
    assert 'train' not in test_scopes['1']


def test_train_split():

    df = get_fake_dataset()
    tr_df, test_df = df.train_split(subjects=[0])
    assert tr_df.shape == (1, 5)
    assert 0 in tr_df.index
    assert test_df.shape == (2, 5)
    assert 1 in test_df.index
    assert 2 in test_df.index


def get_fake_multi_index_dataset():

    fake = Dataset()
    fake['0'] = [1, 2, 3, 4, 5, 6]
    fake['1'] = [1, 2, 3, 4, 5, 6]
    fake['2'] = [1, 2, 3, 4, 5, np.nan]
    fake['subj'] = ['s1', 's2', 's3', 's1', 's2', 's3']
    fake['event'] = ['e1', 'e1', 'e1', 'e2', 'e2', 'e2']
    fake.set_index(['subj', 'event'], inplace=True)

    return fake


def test_multi_index_proc_cv_strategy():

    df = get_fake_multi_index_dataset()

    cv_params = CVStrategy(train_only_subjects=['s1'])
    cv = df._proc_cv_strategy(cv_params)

    assert len(cv.train_only) == 2
    assert df.loc[cv.train_only].shape == (2, 3)

    subjects = df.get_subjects('all', return_as='flat index')
    assert len(subjects) == 6

    # Make sure flat index works for train only internally
    _, subjects, train_only = cv.get_train_only(subjects)

    assert len(subjects) == 4
    assert len(train_only) == 2


def test_multi_index_proc_cv_strategy_groups():

    df = get_fake_multi_index_dataset()
    df = df.copy_as_non_input('0', 'zero')

    cv_params = CVStrategy(groups='zero')
    cv = df._proc_cv_strategy(cv_params)

    assert len(cv.groups) == 6
    assert cv.groups.nunique() == 6


def test_multi_index_set_test_split():

    df = get_fake_multi_index_dataset()
    df.verbose = -1

    df = df.set_test_split(size=1, cv_strategy=None, random_state=None)
    assert len(df.test_subjects) == 1
    assert len(df.train_subjects) == 5

    df = df.set_test_split(size=0, cv_strategy=None, random_state=None)
    assert len(df.test_subjects) == 0
    assert len(df.train_subjects) == 6

    df = df.set_test_split(size=.3, cv_strategy=None, random_state=None)
    assert len(df.test_subjects) == 2
    assert len(df.train_subjects) == 4

    df.set_test_split(size=.5,
                      cv_strategy=CVStrategy(train_only_subjects=['s1']),
                      random_state=1, inplace=True)

    assert len(df.test_subjects) == 2
    assert len(df.train_subjects) == 4
    assert df.loc[df.train_subjects].shape == (4, 3)
    assert ('s1', 'e1') in df.train_subjects

    temp_loc = os.path.join(tempfile.gettempdir(), 'temp.txt')
    df.save_test_split(temp_loc)
    df = df.set_test_split(subjects=temp_loc)
    os.remove(temp_loc)

    assert len(df.test_subjects) == 2
    assert len(df.train_subjects) == 4
    assert df.loc[df.train_subjects].shape == (4, 3)
    assert ('s1', 'e1') in df.train_subjects


def test_multi_index_set_train_split():

    df = get_fake_multi_index_dataset()

    df = df.set_train_split(size=1, cv_strategy=None, random_state=None)
    assert len(df.test_subjects) == 5
    assert len(df.train_subjects) == 1

    df = df.set_train_split(size=.2, cv_strategy=None, random_state=None)
    assert len(df.test_subjects) == 5
    assert len(df.train_subjects) == 1

    df.set_train_split(size=2,
                       cv_strategy=CVStrategy(train_only_subjects=['s1']),
                       random_state=1, inplace=True)

    assert len(df.test_subjects) == 2
    assert len(df.train_subjects) == 4
    assert df.loc[df.train_subjects].shape == (4, 3)
    assert ('s1', 'e1') in df.train_subjects

    temp_loc = os.path.join(tempfile.gettempdir(), 'temp.txt')
    df.save_train_split(temp_loc)
    df = df.set_train_split(subjects=temp_loc)
    os.remove(temp_loc)

    assert len(df.test_subjects) == 2
    assert len(df.train_subjects) == 4
    assert df.loc[df.train_subjects].shape == (4, 3)
    assert ('s1', 'e1') in df.train_subjects



import numpy as np
import pandas as pd
import tempfile
import os
import pytest
from ...main.input_operations import ValueSubset, Intersection
from .datasets import (get_fake_dataset, get_fake_dataset2, get_full_dataset,
                       get_fake_multi_index_dataset, get_full_int_index_dataset,
                       get_fake_dataset4)


def test_get_subjects_None():

    df = get_fake_dataset()

    subjects = df.get_subjects(None, return_as='set')
    assert len(subjects) == 0

    subjects = df.get_subjects(None, return_as='index')
    assert len(subjects) == 0

    subjects = df.get_subjects(None, return_as='flat index')
    assert len(subjects) == 0


def test_get_subjects_tr_test_case():

    df = get_full_dataset()

    assert len(df.get_subjects('all')) == 5
    assert len(df.get_subjects('train')) == 3
    assert len(df.get_subjects('test')) == 2

    # Make sure updates to dropped subjects
    df = df.drop('s1')
    assert len(df.get_subjects('train')) == 2
    df = df.drop('s4')
    assert len(df.get_subjects('test')) == 1


def test_get_subjects_int_index_tr_test_case():

    df = get_full_int_index_dataset()

    assert len(df.get_subjects('all')) == 5
    assert len(df.get_subjects('train')) == 3
    assert len(df.get_subjects('test')) == 2

    # Make sure updates to dropped subjects
    df = df.drop(1)
    assert len(df.get_subjects('train')) == 2
    df = df.drop(4)
    assert len(df.get_subjects('test')) == 1


def test_get_subjects_intersection():

    df = get_fake_dataset4()
    assert len(df.get_subjects('all')) == 6
    assert len(df.get_subjects([1, 2])) == 2

    subjs = df.get_subjects(Intersection([[1, 2]]))
    assert len(subjs) == 2

    subjs = df.get_subjects(Intersection([[1, 2], [2, 3]]))
    assert len(subjs) == 1

    subjs = df.get_subjects(Intersection([[1, 2], [2, 3], [4, 5]]))
    assert len(subjs) == 0

    subjs = df.get_subjects(Intersection([[1, 2], [2, 3], [2, 3]]))
    assert len(subjs) == 1


def test_get_subjects_nan():

    # Subject id == 0 is only one with NaN
    df = get_fake_dataset()
    index_dtype = df.index.dtype.name

    subjects = df.get_subjects('nan', return_as='set')
    assert len(subjects) == 1
    assert df.loc[subjects].shape == (1, 3)
    subj = subjects.pop()
    assert subj == 0
    assert isinstance(subj, int)

    subjects = df.get_subjects('nan', return_as='flat index')
    assert len(subjects) == 1
    assert df.loc[subjects].shape == (1, 3)
    assert subjects[0] == 0
    assert subjects[0].dtype == index_dtype

    subjects = df.get_subjects('nan', return_as='index')
    assert len(subjects) == 1
    assert df.loc[subjects].shape == (1, 3)
    assert subjects[0] == 0
    assert subjects[0].dtype == index_dtype

    # Dataset with no Nan's
    df = get_fake_dataset2()
    subjects = df.get_subjects('nan', return_as='set')
    assert len(subjects) == 0


def test_get_subjects_ValueSubset():

    df = get_fake_dataset()

    # Int column test 1 value
    vs = ValueSubset(name='1', values=1, decode_values=False)
    subjects = df.get_subjects(vs, return_as='set')
    assert len(subjects) == 1
    assert subjects.pop() == 0

    # Int column test 2 values
    vs = ValueSubset(name='1', values=[1, 2], decode_values=False)
    subjects = df.get_subjects(vs, return_as='set')
    assert len(subjects) == 2

    # Str column test 1 value
    vs = ValueSubset(name='2', values=['6'], decode_values=False)
    subjects = df.get_subjects(vs, return_as='set')
    assert len(subjects) == 1
    assert subjects.pop() == 0

    # Str column test 2 values
    vs = ValueSubset(name='2', values=['6', '7'], decode_values=False)
    subjects = df.get_subjects(vs, return_as='set')
    assert len(subjects) == 2

    # Str / cat column test extra values
    vs = ValueSubset(name='2', values=['6', '7', '9'], decode_values=False)
    subjects = df.get_subjects(vs, return_as='set')
    assert len(subjects) == 2

    # Column w/ nan
    vs = ValueSubset(name='3', values=np.nan, decode_values=False)
    subjects = df.get_subjects(vs, return_as='set')
    assert len(subjects) == 1
    assert subjects.pop() == 0

    # Column w/ nan 2 values
    vs = ValueSubset(name='3', values=[np.nan, 2], decode_values=False)
    subjects = df.get_subjects(vs, return_as='set')
    assert len(subjects) == 2
    assert df.loc[subjects].shape == (2, 3)

    # Bad name col
    vs = ValueSubset(name=1, values=[np.nan, 2], decode_values=False)
    with pytest.raises(KeyError):
        subjects = df.get_subjects(vs, return_as='set')

    # Bad input
    with pytest.raises(ValueError):
        vs = ValueSubset(name=[1, 2], values=2)


def test_get_subjects_base():

    df = get_fake_dataset()

    subjs = [0]
    subjects = df.get_subjects(subjs, return_as='set')
    assert len(subjects) == 1
    assert df.loc[subjects].shape == (1, 3)

    subjs = ['0']
    subjects = df.get_subjects(subjs, return_as='set')
    assert len(subjects) == 1
    assert df.loc[subjects].shape == (1, 3)

    subjs = np.array([0])
    subjects = df.get_subjects(subjs, return_as='set')
    assert len(subjects) == 1
    assert df.loc[subjects].shape == (1, 3)

    subjs = pd.Index(data=[0], name=df.index.name)
    subjects = df.get_subjects(subjs, return_as='set')
    assert len(subjects) == 1
    assert df.loc[subjects].shape == (1, 3)

    subjs = pd.Index(data=np.array([0, 2]), name=df.index.name)
    subjects = df.get_subjects(subjs, return_as='index')
    assert len(subjects) == 2
    assert df.loc[subjects].shape == (2, 3)
    assert np.array_equal(np.array([0, 2]), subjects)

    subjects = df.get_subjects(subjs, return_as='flat index')
    assert len(subjects) == 2
    assert df.loc[subjects].shape == (2, 3)
    assert np.array_equal(np.array([0, 2]), subjects)


def test_get_subjects_base_file():

    df = get_fake_dataset()

    temp_loc = os.path.join(tempfile.gettempdir(), 'temp.txt')
    with open(temp_loc, 'w') as f:
        f.write('0\n')

    subjects = df.get_subjects(temp_loc, return_as='set')
    assert len(subjects) == 1
    assert df.loc[subjects].shape == (1, 3)

    temp_loc = os.path.join(tempfile.gettempdir(), 'temp.txt')
    with open(temp_loc, 'w') as f:
        f.write('0\n')
        f.write('1\n')

    subjects = df.get_subjects(temp_loc, return_as='set')
    assert len(subjects) == 2
    assert df.loc[subjects].shape == (2, 3)


def test_multi_index_get_subjects():

    df = get_fake_multi_index_dataset()

    subjs = df.get_subjects(subjects=None)
    assert len(subjs) == 0

    subjs = df.get_subjects(subjects=['s1'])
    assert len(subjs) == 2
    assert ('s1', 'e1') in subjs
    assert ('s1', 'e2') in subjs

    subjs = df.get_subjects(subjects=['s1'], only_level=0)
    assert len(subjs) == 1
    assert 's1' in subjs

    subjs = df.get_subjects(subjects=['s1'], return_as='index',
                            only_level=0)
    assert 's1' in subjs
    assert subjs.name == 'subj'

    subjs = df.get_subjects(subjects=['s1'], return_as='index',
                            only_level=None)
    assert ('s1', 'e1') in subjs
    assert ('s1', 'e2') in subjs
    assert subjs.names == ['subj', 'event']

    subjs = df.get_subjects(subjects=(['s1', 's2'], ['e1']),
                            return_as='set')

    assert len(subjs) == 2
    assert ('s1', 'e1') in subjs
    assert ('s2', 'e1') in subjs

    subjs = df.get_subjects(subjects=('all', ['e1']),
                            return_as='set')
    assert len(subjs) == 3
    assert df.loc[subjs].shape == (3, 3)



import pytest
from ..input_operations import Select
from ..input import Model, Scaler
from ...util import BPtInputMixIn


def test_select_fail():

    s = Select([Model('dt'), Scaler('robust')])

    with pytest.raises(RuntimeError):
        s._check_args()


def test_select_constructor():

    s = Select([Model('dt'), Model('ridge')])
    assert s._constructor == Model._constructor


def test_select_attributes():

    s = Select([Model('dt'), Model('ridge')])
    assert isinstance(s, list)
    assert isinstance(s, Select)
    assert isinstance(s, BPtInputMixIn)


from setuptools import setup, find_packages

setup(name='brain-pred-toolbox',
      version='2.0.0',
      description='The Brain Predictability toolbox (BPt) is a ' +
      'Python based machine learning library designed to work with ' +
      'a range of neuroimaging data.',
      url='http://github.com/sahahn/BPt',
      author='Sage Hahn',
      author_email='sahahn@euvm.edu',
      license='MIT',
      packages=find_packages(),
      python_requires=">=3.7",
      install_requires=[
          'scikit-learn>=0.23.1',
          'numpy==1.19.5',
          'scipy>=1.2',
          'pandas>=1.1.5',
          'matplotlib>=3.2.2',
          'seaborn>=0.9',
          'scikit-image>=0.16',
          'tqdm>=4.51',
          'nevergrad==0.4.3',
          'Ipython',
          'joblib>=1',
          'loky',
      ],
      extras_require={
          'extra': ['lightgbm>3',
                    'sweetviz>2',
                    'nilearn>=0.7',
                    'python-docx',
                    'mvlearn',
                    'imblearn',
                    'networkx',
                    'xgboost>=1.3'],
       },
      test_suite='pytest',
      tests_require=['pytest', 'coverage'],
      zip_safe=False)



import pandas as pd
import numpy as np
from .data_file import load_data_file_proxy
from copy import deepcopy
from .helpers import verbose_print
from pandas.util._decorators import doc
import pandas.core.common as com
from pandas.core.dtypes.generic import ABCMultiIndex


_shared_docs = {}
_shared_docs['scope'] = """scope : :ref:`Scope`
            A BPt style :ref:`Scope` used to select a subset of
            column(s) in which to apply the current function to.
            See :ref:`Scope` for more information on how this
            can be applied.
"""

_shared_docs['inplace'] = """inplace : bool, optional
            If True, perform the current function inplace and return None.

            ::

                default = False

"""
_sip_docs = _shared_docs.copy()

_shared_docs['load_func'] = """load_func : python function, optional
            | Fundamentally columns of type 'data file' represent
              a path to a saved file, which means you must
              also provide some information on how to load the saved file.
              This parameter is where that loading function should be passed.
              The passed `load_func` will be called on each file individually
              and whatever the output of the function is will be passed to
              the different loading functions.

            | You might need to pass a user defined custom function
              in some cases, e.g., you want to use :func:`numpy.load`,
              but then also :func:`numpy.stack`.
              Just wrap those two functions in one,
              and pass the new function.

            ::

                def my_wrapper(x):
                    return np.stack(np.load(x))

            | Note that in this case where a custom function is defined
              it is reccomended that you define this function in
              a separate file from where the main script will be run,
              and then import the function.

            | By default this function assumes data files are passed
              as numpy arrays, and uses the default function
              :func:`numpy.load`, when nothing else is specified.

            ::

                default = np.load

"""

_file_docs = {}
_file_docs['reduce_func'] = '''reduce_func : python function, optional
            The passed python function will be applied only if
            the requested col/column is a 'data file'. In the case
            that it is, the function should accept as input
            the data from one data file, and should return a single
            scalar value. For example, the default value is
            numpy's mean function, which returns one value.

            ::

                default = np.mean

'''

_file_docs['n_jobs'] = '''n_jobs : int, optional
            As with reduce_func, this parameter is only
            valid when the passed col/column is a 'data file'.
            In that case, this specifies the number of cores
            to use in loading and applying the reduce_func to each
            data file. This can provide a significant speed up when
            passed the number of avaliable cores, but can sometimes
            be memory intensive depending on the underlying size of the file.

            If set to -1, will try to automatically use all avaliable cores.

            ::

                default = -1

'''

_shared_docs['drop'] = '''drop : bool, optional
        By default this function will drop any subjects / index that are
        determined to be outliers. On the otherhand, you
        may instead set specific outlier values as NaN values instead.
        To do this, set drop=False. Now those specific
        values identified as outliers will be replaced with NaN.

        ::

            default = True

'''

_shared_docs['subjects'] = '''subjects : :ref:`Subjects`
        This argument can be any of the BPt accepted
        subject style inputs. E.g., None, 'nan' for subjects
        with any nan data, the str location of a file
        formatted with one subject per line, or directly an
        array-like of subjects, to name some options.

        See :ref:`Subjects` for all options.
'''


class Dataset(pd.DataFrame):
    '''| The BPt Dataset class is the main class used for preparing data
      into a compatible format to work with machine learning. This class is new
      as of BPt version 2 (replacing the building in loading functions of the
      old BPt_ML).

    | See :ref:`loading_data` for more a comprehensive guide on this object.

    | This class can be initialized like a pandas.DataFrame, or
      typically from a pandas.DataFrame. This class has some constraints
      relative to using DataFrames. Some of these are that columns must be
      strings (if passed as int-like will be cast to strings), and that
      there cannot be duplicate column names.

    | This class can be initialized in most of the same
      ways that a pandas DataFrame can be initialized, for example

    .. ipython:: python

        data = bp.Dataset(np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
                          columns=['a', 'b', 'c'])
        data

    Or from a pandas DataFrame.

    .. ipython:: python

        import pandas as pd
        df = pd.DataFrame([1, 2, 3], columns=['a'])
        df

        data = bp.Dataset(df)
        data

    The Dataset also has some extra optional constructor parameters:
    roles, scopes, targets and non_inputs, which are just helpers
    for setting parameters at the time of construction. For example:

    .. ipython:: python

        data = bp.Dataset([1, 2, 3], columns=['1'], targets=['1'])
        data.get_roles()

    .. versionadded:: 2.0.0

    '''

    _metadata = ['roles', 'scopes', 'encoders', 'file_mapping',
                 'verbose_', 'test_subjects', 'train_subjects']

    _print = verbose_print

    def __init__(self, data=None, index=None,
                 columns=None, dtype=None,
                 copy=None, roles=None,
                 scopes=None, targets=None,
                 non_inputs=None):

        super().__init__(data=data, index=index,
                         columns=columns, dtype=dtype,
                         copy=copy)

        if roles is not None:
            self.roles = roles
            self._check_roles()

        if scopes is not None:
            self.scopes = scopes
            self._check_scopes()

        if targets is not None:
            self.set_target(targets, inplace=True)
            self._check_roles()

        if non_inputs is not None:
            self.set_non_input(non_inputs, inplace=True)
            self._check_roles()

    def __getitem__(self, key):

        try:
            item = super().__getitem__(key)

        # If the original passed index isn't a key,
        # then try again treating it as a scope.
        except KeyError:
            item = super().__getitem__(self.get_cols(key))

        # If index returns Dataset, copy over metadata
        if isinstance(item, Dataset):
            self._copy_meta_data(item)

        return item

    @property
    def _constructor(self):
        return Dataset

    @property
    def reserved_roles(self):
        '''The dataset class has three fixed and therefore
        reserved (i.e., cannot be set as a scope) roles,
        these are ::

            ['data', 'target', 'non input']

        '''
        return set(['data', 'target', 'non input'])

    @property
    def reservered_scopes(self):
        '''There are a number of reserved fixed scopes
        in the Dataset class, these are ::

            ['all', 'float', 'category', 'data float',
             'data', 'data file', 'data category',
             'non input', 'target', 'target category',
             'target float']
        '''
        return set(['all', 'float', 'category', 'data float',
                    'data', 'data file', 'data category',
                    'non input', 'target', 'target category',
                    'target float'])

    @property
    def verbose(self):
        '''This parameter takes a verbosity level
        as an integer, where level 0 is just warnings,
        a value lower than 0 can be set to mute warnings
        and then higher values for more verbosity.
        E.g., level 1 will print basic information,
        where level 2 or higher will print more extensive information.'''

        if not hasattr(self, 'verbose_'):
            self.verbose_ = 0
        if self.verbose_ is None:
            self.verbose_ = 0

        return self.verbose_

    @verbose.setter
    def verbose(self, verbose):
        self.verbose_ = verbose

    def _inplace(self, func_name, args):
        '''Assumes that inplace is True for this func to be called.'''

        # Create shallow copy
        data_copy = self.copy(deep=False)

        # Get args to pass to copy
        copy_args = {key: args[key] for key in args
                     if key != 'self' and key != 'inplace'}
        copy_args['inplace'] = True

        # Apply func in place
        getattr(data_copy, func_name)(**copy_args)

        # Return copy
        return data_copy

    def _check_test_subjects(self):

        if not hasattr(self, 'test_subjects'):
            self.test_subjects = None

    def _check_train_subjects(self):

        if not hasattr(self, 'train_subjects'):
            self.train_subjects = None

    def _check_file_mapping(self):

        if not hasattr(self, 'file_mapping'):
            self.file_mapping = {}
        elif getattr(self, 'file_mapping') is None:
            self.file_mapping = {}

        # Check to remove any non relevant file mappings
        # e.g., ones that have been dropped
        data_file_cols = self.get_cols(scope='data file')

        # Get all unique files, including if any NaNs
        unique_files = pd.unique(self[data_file_cols].values.ravel())

        # If any NaN's, make sure np.nan is in the file_mapping
        if pd.isnull(unique_files).any():
            self.file_mapping[np.nan] = np.nan

        # Compute the difference
        file_mapping_keys = np.array(list(self.file_mapping))

        # Make sure types are lined up
        if file_mapping_keys.dtype.name != unique_files.dtype.name:
            unique_files = unique_files.astype(file_mapping_keys.dtype.name)

        # Compute difference
        to_remove = np.setdiff1d(file_mapping_keys, unique_files,
                                 assume_unique=True)

        # Remove any difference
        for key in to_remove:
            try:
                self.file_mapping.pop(key)
            except KeyError:
                pass

    def _check_encoders(self):

        if not hasattr(self, 'encoders'):
            self.encoders = {}
        elif getattr(self, 'encoders') is None:
            self.encoders = {}

    def _check_cols(self):
        '''BPt / scopes assume that all columns are of
        type str. This function should be called before
        check roles or check scope. Also checks to make sure
        no columns have one of the reserved role names'''

        col_names = list(self)

        # Check if any col names not str's
        non_str = []
        for col in col_names:
            if not isinstance(col, str):
                non_str.append(col)

        if len(non_str) > 0:

            rename = {c: str(c) for c in non_str}
            super().rename(rename, axis=1, inplace=True)

            self._print('Warning: the columns:', repr(non_str),
                        'were cast to str', level=0)

        # Check if any columns have the same name as a reserved scope
        reserved_overlap = set(self.columns).intersection(
            self.reservered_scopes)
        if len(reserved_overlap) > 0:
            raise RuntimeError('The columns: ' + repr(reserved_overlap) + ' '
                               'overlap with reserved saved names! '
                               'They must be changed.')

        # Check to make sure there are no columns with the same name,
        # i.e., duplicates
        self._check_duplicate_cols()

    def _check_duplicate_cols(self):

        if len(set(self.columns)) != len(self.columns):
            raise RuntimeError('Duplicate columns with the same name ',
                               'are not allowed!')

    def _check_roles(self, check_type=True):

        # Make sure cols type str
        if check_type:
            self._check_cols()

        if not hasattr(self, 'roles'):
            self.roles = {}
        elif getattr(self, 'roles') is None:
            self.roles = {}

        if not isinstance(self.roles, dict):
            raise RuntimeError('roles must be a dict.')

        # Fill in any column without a role.
        for col in list(self.columns):
            if col not in self.roles:
                self._set_role(col, 'data')

        # Remove any saved roles that are not
        # in the data any more.
        for col in list(self.roles):
            if col not in self.columns:
                del self.roles[col]

    def _check_sr(self):
        '''Helper to check scopes and roles'''

        # Only check once when both being checked
        self._check_cols()

        # Check scopes and roles
        self._check_scopes(check_type=False)
        self._check_roles(check_type=False)

    def _check_scopes(self, check_type=True):

        # Make sure cols type str
        if check_type:
            self._check_cols()

        # If doesn't exist, create scopes
        if not hasattr(self, 'scopes'):
            self.scopes = {}

        # Or is set to None
        elif getattr(self, 'scopes') is None:
            self.scopes = {}

        if not isinstance(self.scopes, dict):
            raise RuntimeError('scopes must be a dict.')

        # Make sure each col is init'ed with a scope
        for col in self.columns:
            if col not in self.scopes:
                self.scopes[col] = set()

        # Make sure to get rid of any removed cols
        for col in list(self.scopes):
            if col not in self.columns:
                del self.scopes[col]

        # @TODO maybe optimize with self.dtypes ?

        # Compute columns which are categorical + columns with scope category
        dtype_category = set(self.select_dtypes('category'))
        scope_category = set([col for col in self.columns if
                             'category' in self.scopes[col]])

        # End if equal
        if dtype_category == scope_category:
            return

        # Add scope to columns which have dtype category but not scope category
        needs_scope = dtype_category - scope_category
        for col in needs_scope:
            self._add_scope(col, 'category')

        # For any columns which have scope category, but the dtype isn't
        # assume that the scope should be removed
        remove_scope = scope_category - dtype_category
        for col in remove_scope:
            self._remove_scope(col, 'category')

    def get_roles(self):
        ''' This function can be
        used to get a dictionary with the currently
        loaded roles, See :ref:`Role` for more information
        on how roles are defined and used within BPt.

        Returns
        --------
        roles : dict
            Returns the up to date dictionary
            of roles stored in self.roles, where
            each key is a column name and each value
            is the corresponding role.
        '''

        self._check_roles()
        return self.roles

    @doc(**_sip_docs)
    def set_role(self, scope, role, inplace=False):
        '''This method is used to set a role for
        either a single column or multiple, as set
        through the scope parameter. See :ref:`Role`
        for more information about how roles are
        used within BPt.

        Parameters
        ----------
        {scope}

        role : :ref:`Role`
            A valid role in which to set all columns
            in the passed scope to. Input must be
            either 'data', 'target' or 'non input'.
            Note: by default all columns start with role 'data'.

            See :ref:`Role` for more information on how
            each role differs.

        {inplace}

        See Also
        ----------
        set_target : Specifically for setting target role.
        set_non_input : Specifically for setting non input role.
        get_roles : Returns a dictionary with saved roles.

        Examples
        ---------
        Setting columns role's within the :class:`Dataset` is an
        essential part of using the object.

        .. ipython:: python

            data = bp.read_csv('data/example1.csv')
            data = data.set_role('animals', 'target')
            data
            data.get_roles()

        We can also use the method to set columns to role non input,
        which has the additional constraint that no NaN values
        can be present in that column. So we can see below
        that one row is dropped.

        .. ipython:: python

            data = data.set_role('numbers', 'non input')
            data
            data.get_roles()

        '''

        if not inplace:
            return self._inplace('set_role', locals())

        self._check_sr()

        cols = self._get_cols(scope)
        for col in cols:
            self._set_role(col, role)

    def set_roles(self, scopes_to_roles, inplace=False):
        '''This method is used to set multiple roles
        across multiple scopes as specified by a passed
        dictionary with keys as scopes and values as
        the role to set for all columns corresponding to that
        scope. See :ref:`Role`
        for more information about how roles are
        used within BPt.

        Parameters
        -----------
        scope_to_roles : dict of :ref:`Scope` to :ref:`Role`
            A python dictionary with keys as :ref:`Scope`
            and their corresponding value's as the :ref:`Role`
            in which those columns should take.

            For example ::

                scope_to_roles = dict()
                scope_to_roles['col1'] = 'target'
                scope_to_roles['col2'] = 'non input'

        inplace : bool, optional
            If True, do operation inplace and return None.

            ::

                default = False

        '''

        if not inplace:
            return self._inplace('set_roles', locals())

        # Init checks just once
        self._check_sr()

        # Set for each passed scope - all in place
        for scope, role in zip(scopes_to_roles, scopes_to_roles.values()):
            [self._set_role(col, role) for col in self._get_cols(scope)]

    def set_target(self, scope, inplace=False):
        '''This method is used to set
        either a single column, or multiple, specifically
        with role `target`. This function is simply
        a helper wrapper around :func:`Dataset.set_role`.

        See :ref:`Role` for more information
        about how roles are used within BPt.

        Parameters
        ----------
        scope : :ref:`Scope`
            A BPt style :ref:`Scope` used to select a subset of
            columns in which to set with role `target`.

        inplace : bool, optional
            If True, do operation inplace and return None.

            ::

                default = False
        '''

        return self.set_role(scope=scope, role='target', inplace=inplace)

    def set_non_input(self, scope, inplace=False):
        '''This method is used to set
        either a single column, or multiple, specifically
        with role `non input`. This function is simply
        a helper wrapper around :func:`Dataset.set_role`.

        See :ref:`Role` for more information
        about how roles are used within BPt.

        Parameters
        ----------
        scope : :ref:`Scope`
            A BPt style :ref:`Scope` used to select a subset of
            columns in which to set with role `non input`.

        inplace : bool, optional
            If True, do operation inplace and return None.

            ::

                default = False
        '''

        return self.set_role(scope=scope, role='non input', inplace=inplace)

    def _set_role(self, col, role):
        '''Internal function for setting a single columns role.'''

        if role not in self.reserved_roles:
            raise AttributeError(
                'Passed role "' + str(role) + '" must be one of ' +
                str(self.reserved_roles))

        # Set as role
        self.roles[col] = role

        # If role is non input, can't have any NaN
        if role == 'non input':
            self.drop_nan_subjects(scope=col, inplace=True)

    def get_scopes(self):
        '''This returns the up to date scopes for the
        Dataset. Note that scopes can be accessed through
        self.scopes, but this function is reccomended instead,
        as self.scope may be missing the latest change applied.

        Returns
        --------
        scopes : dict of sets
            Returns a dictionary with column name as the
            key and a set containing all scopes associated with
            each column.
        '''

        self._check_scopes()
        return self.scopes

    def add_scope(self, scope, scope_val, inplace=False):
        '''This method is designed as helper for adding a new scope val
        to a number of columns at once, using the existing scope system.
        Don't be confused about the arguments, the scope parameter is used
        to select the columns in which the scope_val should be added as a
        scope to those columns.

        See :ref:`Scope` for more information
        on how scope's are used in BPt. This
        function specifically is for adding
        custom, tag-like, scopes to certain columns.

        Parameters
        -----------
        scope : :ref:`Scope`
            A BPt style :ref:`Scope` used to select a subset of
            column in which to add new scopes (scope_val) to.

        scope_val : str or array-like of str
            A single string scope value to add
            to the columns select by scope, or an array-like of
            scope values to all add to the selected
            col. E.g.,

            ::

                scope_val = '1'

            Would add '1' as a scope

            ::

                scope_val = ['1', '2']

            Would add '1' and '2' as scopes.

        inplace : bool, optional
            If True, do operation inplace and return None.

            ::

                default = False
        '''

        if not inplace:
            return self._inplace('add_scope', locals())

        self._check_sr()

        cols = self._get_cols(scope)
        for col in cols:
            self._add_scope(col, scope_val)

    def _add_scope(self, col, scope_val):

        # Check if category
        if scope_val == 'category':
            self.scopes[col].add(scope_val)

            # If already categorical, skip
            if self[col].dtype.name == 'category':
                return

            # Special case, if converting from bool to category
            if self[col].dtype.name == 'bool':

                # Need to cast to int first as intermediate
                self[col] = self[col].astype('int')

            new_col = self[col].copy().astype('category')
            new_col.cat.as_ordered(inplace=True)
            self[col] = new_col

            return

        # If a int or float
        if isinstance(scope_val, int) or isinstance(scope_val, float):
            scope_val = str(scope_val)

            # @TODO Add a verbose here saying cast to str?

        # If str
        if isinstance(scope_val, str):

            # Make sure doesn't overlap with loaded column
            if scope_val in list(self):
                raise RuntimeError('Warning scope of: ' + scope_val + ' '
                                   'overlaps with a loaded column. This '
                                   'can cause index errors, as if this '
                                   'scope is '
                                   'requested, then the column will be '
                                   'returned '
                                   'instead of this scope!')

            # Add
            self.scopes[col].add(scope_val)

        # Or assume array-like
        else:
            for s in set(list(scope_val)):
                self._add_scope(col, s)

    def remove_scope(self, scope, scope_val, inplace=False):
        '''This method is used for removing scopes
        from an existing column or subset of columns,
        as selected by
        the scope parameter.

        See :ref:`Scope` for more information
        on how scope's are used in BPt. This
        function specifically is for removing
        custom, tag-like, scopes to certain columns.

        See related method :func:`add_scope <Dataset.add_scope>`

        Parameters
        -----------
        scope : :ref:`Scope`
            A BPt style :ref:`Scope` used to select a subset of
            column in which to remove scopes (scope_val).

        scope_val : str or array-like of str
            A single string scope value to remove
            from the column(s), or an array-like of
            scope values to all remove from the selected
            column(s).

        inplace : bool, optional
            If True, do operation inplace and return None.

            ::

                default = False
        '''

        if not inplace:
            return self._inplace('remove_scope', locals())

        self._check_sr()

        cols = self._get_cols(scope)
        for col in cols:
            self._remove_scope(col, scope_val)

    def _remove_scope(self, col, scope_val):

        try:
            self.scopes[col].remove(scope_val)

            # If removing category and currently pandas dtype is category,
            # change to float32.
            if scope_val == 'category' and \
               self[col].dtype.name == 'category':
                self[col] = self[col].astype('float32')

        except KeyError:
            pass

    def _get_reserved_cols(self, scope, columns):
        '''If passed here then we know the
        passed scope is one the options.'''

        if scope == 'all':
            return list(columns)

        elif scope == 'data float':
            return [col for col in columns if
                    'category' not in self.scopes[col] and
                    self.roles[col] == 'data']

        elif scope == 'target float':
            return [col for col in columns if
                    'category' not in self.scopes[col] and
                    self.roles[col] == 'target']

        elif scope == 'float':
            return [col for col in columns if
                    'category' not in self.scopes[col]]

        elif scope == 'category':
            return [col for col in columns if
                    'category' in self.scopes[col]]

        elif scope == 'data category':
            return [col for col in columns if
                    'category' in self.scopes[col] and
                    self.roles[col] == 'data']

        elif scope == 'target category':
            return [col for col in columns if
                    'category' in self.scopes[col] and
                    self.roles[col] == 'target']

        elif scope == 'data':
            return [col for col in columns if
                    self.roles[col] == 'data']

        elif scope == 'target':
            return [col for col in columns if
                    self.roles[col] == 'target']

        elif scope == 'non input':
            return [col for col in columns if
                    self.roles[col] == 'non input']

        elif scope == 'data file':
            return [col for col in columns if
                    'data file' in self.scopes[col]]

        else:
            raise RuntimeError('Should be impossible to reach here.')

    def _g_cols(self, scope, columns, saved_scopes):
        '''This method is used internally to avoid
        some repeated actions.

        It will return columns as a list.'''

        # If int or float, cast to str.
        if isinstance(scope, int) or isinstance(scope, float):
            scope = str(scope)

        # Check is passed scope is reserved
        if isinstance(scope, str):

            # First case, check if reserved scope
            if scope in self.reservered_scopes:
                return self._get_reserved_cols(scope, columns)

            # Next case is if the passed scope
            # is a valid column.
            elif scope in columns:
                return [scope]

            # Check if a saved scope
            if scope in saved_scopes:
                return [col for col in columns if
                        scope in self.scopes[col]]

            # Last string case is a stub search
            return [col for col in columns if scope in col]

        # If scope not str, assume iterable, and call recursively
        # on each iterable. If not iterable, will fail
        cols = []
        for scp in scope:
            cols += self._g_cols(scp, columns, saved_scopes)

        # In the case that scope was an iterable, their exists
        # the possibility for overlapping columns.
        # Cast to set and back to list to remove these
        return list(set(cols))

    def _get_cols(self, scope, limit_to=None):

        # Set columns to check within based on if
        # limit_to is passed
        if limit_to is None:
            columns = self.columns
        else:
            columns = self._get_cols(limit_to, limit_to=None)

        # Get the set of saved valid scopes for these columns
        saved_scopes = set()
        for col in columns:
            saved_scopes.update(self.scopes[col])

        # Get columns from internal g_cols func
        cols = self._g_cols(scope, columns, saved_scopes)

        # For maximum reproducibility and reliable behavior
        # always return the sorted columns.
        return sorted(cols)

    def get_cols(self, scope, limit_to=None):
        '''This method is the main internal and external
        facing way of getting the names of columns which match a
        passed scope from the Dataset. For example this
        method may be useful for user's when they want to ensure
        that a scope returns exactly the subset of columns they expect.

        Parameters
        -----------
        scope : :ref:`Scope`
            The BPt style :ref:`Scope` input that will be
            used to determine which column names from the Dataset
            to return.

        limit_to : :ref:`Scope` or None, optional
            Can optionally limit the columns
            to search over with another scope.
            If None, then will use 'all'.

            ::

                default = None

        Returns
        ----------
        cols : list
            This method returns the columns specified by
            the input `scope` argument as a sorted python
            list

        '''

        # Perform checks first
        self._check_sr()

        return self._get_cols(scope=scope, limit_to=limit_to)

    def _get_data_cols(self, ps_scope):
        return self._get_cols('data', limit_to=ps_scope)

    def _get_data_inds(self, ps_scope, scope):
        '''This function always limits first by the data cols,
        then ps_scope refers to the problem_spec scope, and
        lastly scope can be used to specify of subset of those columns'''

        # Get data cols, these are the ordered columns
        # actually passed to an evaluate function
        data_cols = self._get_data_cols(ps_scope)

        # If scope is 'all', data inds
        if scope == 'all':
            return list(range(len(data_cols)))

        # Otherwise, get subset of inds, then return sorted
        inds = [data_cols.index(k) for k in
                self._get_cols(scope, limit_to=data_cols)]

        return sorted(inds)

    def _is_data_cat(self, ps_scope, scope):

        # Get data cols
        data_cols = self._get_data_cols(ps_scope)

        # Check if all are categorical cat
        # In the case where the underlying scope
        # is nothing, return False
        try:
            all_cat = self._is_category(scope, limit_to=data_cols,
                                        check_scopes=False)
        except KeyError:
            all_cat = False

        return all_cat

    @doc(**_file_docs)
    def get_values(self, col, dropna=True, decode_values=False,
                   reduce_func=np.mean, n_jobs=-1):
        '''This method is used to obtain the either normally loaded and
        stored values from a passed column, or in the case of a data file
        column, the data file proxy values will be loaded. There is likewise
        an option to return these values with and without missing values
        included.

        Parameters
        -----------
        col : str
            The name of the column in which to load/extract values for.

        dropna : bool, optional
            Boolean argument, if True, return only non-nan values.
            If False, return everything regardless of if NaN.

            ::

                default = True

        decode_values : bool, optional
            Boolean argument, if True, then try to
            return the original values before any encoding,
            otherwise default of False will return the current
            loaded values.

            ::

                default = False

        {reduce_func}

        {n_jobs}

        Returns
        ---------
        values : pandas Series
            This method returns a single Series with the extracted
            values for the requested column, which either include or
            exclude missing values and may be data file proxy values
            depending on the nature of the requested column.

        '''

        # Check scopes and roles
        self._check_sr()

        return self._get_values(col=col, dropna=dropna,
                                decode_values=decode_values,
                                reduce_func=reduce_func, n_jobs=n_jobs)

    def _get_values(self, col, dropna=True, decode_values=False,
                    reduce_func=np.mean, n_jobs=-1):

        # Extract base series depending on if dropna
        if dropna:
            values = self[col].loc[self._get_non_nan_subjects(col)]
        else:
            values = self[col]

        # If data file, load and return proxy
        if 'data file' in self.scopes[col]:

            # If NaN, make sure NaN is in file mapping
            # if passing any NaN's here
            if pd.isnull(values).any():
                self.file_mapping[np.nan] = np.nan

            return load_data_file_proxy(values,
                                        reduce_func=reduce_func,
                                        file_mapping=self.file_mapping,
                                        n_jobs=n_jobs)

        if not decode_values:
            return values

        # Check for if to try and de-code
        # Note that we will assume data files can't be de-coded
        # Check encoders init'ed
        self._check_encoders()

        # Make copy of values
        values = values.copy()

        try:
            encoder = self.encoders[col]

        # If no encoder, return values as is
        except KeyError:
            return values

        # If dict style encoder - any other cases?
        if isinstance(encoder, dict):
            return values.replace(encoder)

        return values

    def _get_nan_subjects(self, col):
        return self[col][self[col].isna()].index

    def _get_non_nan_subjects(self, col):
        return self[col][~self[col].isna()].index

    def auto_detect_categorical(self, scope='all', obj_thresh=30,
                                all_thresh=None, inplace=False):
        '''This function will attempt to automatically add scope "category" to
        any loaded categorical variables. Note that any columns with pandas
        data type category should already be detected without
        calling this function.

        Default heuristic threshold settings are used by default, by
        they can be changed.

        Note: if any of the conditions are
        met the column will be set to categorical, it is not the case
        that if a single condition is not met, then it won't be categorical.

        Any column with only two unique non-nan values
        is considered binary- and therefore categorical. This is a
        fixed behavior of this function.

        If any data file's within scope, will always treat as not categorical,
        and further will make sure it is not categorical if for some reason
        set as that.

        Parameters
        -----------
        scope : :ref:`Scope`, optional
            Any valid BPt style scope used to select which columns this
            function should operate on. E.g., if known that only
            a subset of columns might be categorical, you could
            specify only this subset to work on.

            By default this is set to 'all', which will check all columns.

            ::

                default = 'all'

        obj_thresh : int or None, optional
            This threshold is used for any columns of pandas
            datatype object. If the number of unique non-nan values in
            this object datatype column is less than this threshold,
            this column will be set to categorical.

            To ignore this condition, you may pass None.

            ::

                default = 30

        all_thresh : int or None, optional
            Simmilar to obj_thresh, except that this condition is
            for any column regardless of datatype, this threshold
            is set such that if the number of unique non-nan values
            in this column is less than the passed value, this column
            will be set to categorical.

            To ignore this condition, you may pass None.

            ::

                default = None

        inplace : bool, optional
            If True, do operation inplace and return None.

            ::

                default = False

        '''

        if not inplace:
            return self._inplace('auto_detect_categorical', locals())

        # Check scope and role
        self._check_sr()

        # Get cols from scope
        cols = self._get_cols(scope)

        for col in cols:

            # Skip if data file
            if 'data file' in self.scopes[col]:

                # Make sure not categorical then skip rest of
                # checks.
                self._remove_scope(col, 'category')
                continue

            # Get non-nan values to check
            values = self._get_values(col, dropna=True)

            # First check for binary by unique number of columns
            n_unique = len(values.unique())
            if n_unique == 2:
                self._add_scope(col, 'category')

            # Check object threshold
            if obj_thresh is not None and values.dtype.name == 'object':
                if n_unique < obj_thresh:
                    self._add_scope(col, 'category')

            # Check all threshold
            if all_thresh is not None and n_unique < all_thresh:
                self._add_scope(col, 'category')

        self._print('Num. categorical variables in dataset:',
                    len(self._get_cols(scope='category')), level=1)
        self._print('Categorical variables in dataset:',
                    self._get_cols(scope='category'), level=2)

    def _get_problem_type(self, col):
        '''Return the default problem type for a given column.'''

        self._check_scopes()

        # Doesn't matter if column is categorical
        # if just two unique values, assume binary
        if self[col].nunique(dropna=True) == 2:

            # Further, if not categorical, make sure values
            # are 0 and 1.
            if 'category' not in self.scopes[col]:
                u_values = self.get_values(col, dropna=True).unique()
                if 0 not in u_values or 1 not in u_values:
                    raise RuntimeError(
                        'Error determining default problem type: Requested'
                        ' target column: ' + str(col) + ' has only two unique'
                        ' values: ' + repr(u_values) + ', but they are not'
                        ' categorical and not 0 and 1. Either change the'
                        ' values / type of target column: ' + str(col) + ', '
                        'or pass an explicit option to problem_type'
                        ', e.g., problem_type="binary"')

            self._print('Problem type auto-detected as binary', level=2)
            return 'binary'

        if 'category' in self.scopes[col]:
            self._print('Problem type auto-detected as categorical', level=2)
            return 'categorical'

        self._print('Problem type auto-detected as regression', level=2)
        return 'regression'

    def rename(self, mapper=None, index=None,
               columns=None, axis=None, copy=True,
               inplace=False, level=None, errors='ignore'):
        '''Calls method according to:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.rename.html

        This version is updated to reflect changes to meta-data after renaming, e.g.,
        scopes will be updated to their new name accordingly.

        Parameters
        -------------
        mapper : dict-like or function, optional
            Dict-like or function transformations to apply to that
            axis’ values. Use either mapper and axis to specify
            the axis to target with mapper, or index and columns.

            ::

                default = None

        index : dict-like or function, optional
            Alternative to specifying axis
            (mapper, axis=0 is equivalent to index=mapper).

            ::

                default = None

        columns : dict-like or function, optional
            Alternative to specifying axis
            (mapper, axis=1 is equivalent to columns=mapper).

            ::

                default = None

        axis : {0 or 'index', 1 or 'columns'}, optional
            Axis to target with mapper. Can be either the axis name
            ('index', 'columns') or number (0, 1). The default is ‘index’.

            ::

                default = 0

        copy : bool, optional
            Also copy underlying data.

            ::

                default = True

        inplace : bool, optional
            Whether to return a new :class:`Dataset`.
            If True then value of copy is ignored.

            ::

                default = False

        level : int or level name, optional
            In case of a MultiIndex, only rename labels in the specified level.

            ::

                default = None

        errors : {'ignore', 'raise'}, optional
            If 'raise', raise a KeyError when a dict-like mapper,
            index, or columns contains labels that are not present
            in the Index being transformed. If 'ignore',
            existing keys will be renamed and extra keys will be ignored.

            ::

                default = 'ignore'
        '''

        # Make sure everything up to data
        # and that the attributes exist
        self._check_sr()
        self._check_encoders()
        self._check_train_subjects()
        self._check_test_subjects()

        # First apply parent function
        data = super().rename(mapper=mapper, index=index, columns=columns,
                              axis=axis, copy=copy, inplace=inplace,
                              level=level,
                              errors=errors)

        # Next proc using either result if not inplace
        # or self if inplace
        result = self if inplace else data

        # Set columns and index
        if not (index is not None or columns is not None):
            if axis and self._get_axis_number(axis) == 1:
                columns = mapper
            else:
                index = mapper

        # Process each passed
        for axis_no, replacements in enumerate((index, columns)):
            if replacements is None:
                continue

            # Get rename function
            func = com.get_rename_function(replacements)

            if axis_no == 0:
                result._rename_index(func, level=level)
            elif axis_no == 1:
                result._rename_columns(func)
            else:
                raise RuntimeError(f'axis_no == {axis_no}, should be 0 or 1.')

        return result

    def _rename_columns(self, func):

        # Can't have mapping to duplicate cols
        self._check_duplicate_cols()

        to_change = ['roles', 'scopes', 'encoders']
        for attr in to_change:
            curr_attr = getattr(self, attr)
            new_attr = {func(key): curr_attr[key] for key in curr_attr}
            setattr(self, attr, new_attr)

    def _rename_index(self, func, level=None):

        # Skip if no train or test subjects
        if self.train_subjects is None and self.test_subjects is None:
            return

        if not isinstance(self, ABCMultiIndex):
            to_change = ['test_subjects', 'train_subjects']
            for attr in to_change:
                curr_attr = getattr(self, attr)
                if curr_attr is not None:
                    new_vals = [func(val) for val in curr_attr.values]
                    new_attr = pd.Index(new_vals)
                    setattr(self, attr, new_attr)

        # @TODO Support at some point
        else:
            raise RuntimeError('renaming index in a MultiIndex case '
                               'is not currently supported.')
            '''
            if level is not None:
                level = self.index._get_level_number(level)

            if level is not None:
                items = [
                    tuple(f(y) if i == level else y for i, y in enumerate(x))
                    for x in self
                ]

            else:
                items = [tuple(f(y) for y in x) for x in self]
            '''

    def _is_category(self, scope, limit_to=None, check_scopes=True):
        '''Tests if a set of columns are categorical,
        returns True only if all are categorical.'''

        # Only if requested check scopes
        if check_scopes:
            self._check_scopes()

        # Get cols
        cols = self._get_cols(scope, limit_to=limit_to)

        # If no valid, raise error
        if len(cols) == 0:
            raise KeyError('Passed: ' + repr(scope) +
                           ' not a valid scope.')

        # Check for each col
        is_category = []
        for col in self._get_cols(scope):

            # Only check is there is a scope for this col
            if col in self.scopes:
                is_category.append('category' in self.scopes[col])
            else:
                is_category.append(False)

        return all(is_category)

    def copy(self, deep=True):
        '''Creates and returns a dopy of this dataset, either
        a deep copy or shallow.

        Parameters
        ----------
        deep : bool, optional
            If the returned copy should be
            a deep copy (True), or a shallow copy (False).

            ::

                default == True
        '''

        # Base parent pandas behavior
        dataset = super().copy(deep=deep)

        # Make copy of meta data from self to dataset in place
        # Always makes deep copy for meta-data.
        self._copy_meta_data(dataset)

        return dataset

    def _copy_meta_data(self, dataset):
        '''Add copy of meta data from self to dataset in place'''

        # Make copy of meta data
        for m in self._metadata:

            # If has this attribute, set in the copy a copy.
            if hasattr(self, m):
                setattr(dataset, m, deepcopy(getattr(self, m)))

        return self

    def get_Xy(self, problem_spec='default', **problem_spec_params):
        '''This function is used to get a sklearn-style
        grouping of input data (X) and target data (y)
        from the Dataset as according to a passed problem_spec.

        Note: X and y are returned as pandas DataFrames not Datasets,
        so none of the Dataset meta data is accessible through the
        returned X, y here.

        Parameters
        -----------
        problem_spec : :class:`ProblemSpec` or 'default', optional
            This argument accepts an instance of the
            params class :class:`ProblemSpec`.
            This object is essentially a wrapper around commonly used
            parameters needs to define the context
            the model pipeline should be evaluated in.
            It includes parameters like problem_type, scorer, n_jobs,
            random_state, etc...
            See :class:`ProblemSpec` for more information
            and for how to create an instance of this object.

            If left as 'default', then will initialize a
            ProblemSpec with default params.

            ::

                default = 'default'

        problem_spec_params : :class:`ProblemSpec` params, optional
            You may also pass any valid parameter value pairs here,
            e.g.

            ::

                get_Xy(problem_spec=problem_spec, problem_type 'binary')

            Any parameters passed here will override the original
            value in problem spec. This can be useful when using all default
            values for problem spec except for one, e.g., you just want
            to change random_state.

            ::

                get_Xy(random_state=5)

        Returns
        --------
        X : pandas DataFrame
            DataFrame with the input data and columns as
            specified by the passed problem_spec.

        y : pandas Series
            Series with the the target values as requested
            by the passed problem_spec.
        '''
        from ..main.funcs import problem_spec_check

        # Get proc'ed problem spec
        ps = problem_spec_check(problem_spec, dataset=self,
                                **problem_spec_params)

        # Get sorted subjects from problem spec
        subjects = sorted(self.get_subjects(ps.subjects, return_as='set'))

        # Get X cols
        X_cols = self._get_data_cols(ps.scope)

        # Get X as pandas df subset, y as Series
        X = pd.DataFrame(self.loc[subjects, X_cols]).astype(ps.base_dtype)
        y = self.loc[subjects, ps.target].astype('float64')

        # Add warning for unique n subjs
        n_unique_index = len(pd.unique(X.index))
        if n_unique_index != len(X):
            self._print('The number of unique index rows (',
                        str(n_unique_index),
                        ') does not match the number of rows (',
                        str(len(X)), ').',
                        'There may be duplicate subjects / index.', level=0)

        return X, y

    def _replace_datafiles_with_repr_(self):

        def is_int(i):
            try:
                int(i)
                return True
            except ValueError:
                return False

        for col in self:
            if 'data file' in self.scopes[col]:
                self[col] = [f'Loc({int(i)})' if is_int(i)
                             else repr(i) for i in self[col]]

    def _repr_html_(self):

        # Checks
        self._check_sr()

        template = """<div style="float: left; padding: 10px;">
        <h3>{0}</h3>{1}</div>"""

        html = ''
        for scope in ['data', 'target', 'non input']:
            cols = self._get_cols(scope)

            # Minor formatting for display
            display_scope = scope[0].upper() + scope[1:]
            if display_scope == 'Target' and len(cols) > 1:
                display_scope = 'Targets'
            elif display_scope == 'Non input':
                display_scope = 'Non Input'

            if len(cols) > 0:

                display_copy = self[cols].copy()
                display_copy._replace_datafiles_with_repr_()

                html += template.format(display_scope,
                                        display_copy._base_repr_html_())
                html += '\n'

        return html

    def _base_repr_html_(self):

        html = super()._repr_html_()

        try:
            train_subjects = self.get_subjects('train', return_as='set')
            test_subjects = self.get_subjects('test', return_as='set')
        except RuntimeError:
            return html

        train_color = 'RGBA(176, 224, 230, .25)'
        test_color = 'RGBA(249, 121, 93, .25)'

        train_subjects = set(['>' + str(s) for s in train_subjects])
        test_subjects = set(['>' + str(s) for s in test_subjects])

        # Generate splits 1
        start = html.index('</thead>')
        s1 = html[start:].split('<th')
        for i in range(len(s1)):
            s2 = s1[i].split('</th>')

            for j in range(len(s2)):
                if s2[j] in train_subjects:
                    s2[j] = ' style="background: ' + train_color + '"' + s2[j]
                elif s2[j] in test_subjects:
                    s2[j] = ' style="background: ' + test_color + '"' + s2[j]

            s1[i] = '</th>'.join(s2)

        # Set as re-joined html str
        s1_post = '<th'.join(s1)
        html = html[:start] + s1_post

        n_cols = str(len(self.columns))

        train_info = '<p style="margin-top: .35em"><span style="background: '
        train_info += train_color + '">'
        train_info += str(len(train_subjects)) + " rows × "
        train_info += n_cols + " columns - Train Set </span></p>"

        test_info = '<p style="margin-top: .35em"><span style="background: '
        test_info += test_color + '">'
        test_info += str(len(test_subjects)) + " rows × "
        test_info += n_cols + " columns - Test Set </span></p>"

        extra_info = train_info + test_info

        # Add before end
        html = html[:-6] + extra_info + html[-6:]

        return html

    from ._subjects import (_apply_only_level,
                            _get_nan_loaded_subjects,
                            _get_value_subset_loaded_subjects,
                            _get_base_loaded_subjects,
                            _return_subjects_as,
                            get_subjects)

    from ._data_files import (get_file_mapping,
                              add_data_files,
                              to_data_file,
                              _series_to_data_file,
                              consolidate_data_files,
                              update_data_file_paths,
                              _get_next_ind)

    from ._plotting import (plot,
                            nan_info,
                            _cont_info,
                            _cat_info,
                            summary,
                            plot_bivar,
                            _plot_cat_cat,
                            _plot_cat_float,
                            _plot_float_float,
                            _get_plot_values,
                            _plot_float,
                            _plot_col,
                            _plot_category,
                            _print_plot_info)

    from ._encoding import (_add_new_copy,
                            to_binary,
                            binarize,
                            _base_binarize,
                            _binarize,
                            k_bin,
                            _k_bin,
                            ordinalize,
                            _ordinalize,
                            nan_to_class,
                            copy_as_non_input,
                            add_unique_overlap,
                            _replace_cat_values)

    from ._validation import (_validate_cv_key,
                              _proc_cv_strategy,
                              _validate_split,
                              _finish_split,
                              set_test_split,
                              test_split,
                              train_split,
                              set_train_split,
                              save_test_split,
                              save_train_split)

    from ._filtering import (_drop_subjects,
                             drop_nan_subjects,
                             apply_inclusions,
                             apply_exclusions,
                             drop_subjects_by_nan,
                             _drop_or_nan,
                             _data_file_fail_check,
                             _drop_cols,
                             filter_outliers_by_percent,
                             filter_outliers_by_std,
                             filter_categorical_by_percent,
                             drop_id_cols,
                             drop_duplicate_cols,
                             drop_cols,
                             drop_cols_by_unique_val,
                             drop_cols_by_nan)


def read_csv(*args, **kwargs):
    '''Passes along all arguments and kwargs to
    :func:`pandas.read_csv` then casts to :class:`Dataset`.

    This method is just a helper wrapper function.
    '''

    return Dataset(pd.read_csv(*args, **kwargs))



from .Params import TransitionChoice, Log, Scalar, Array
import numpy as np

P = {}
P['default'] = {}

cls_weight = TransitionChoice([None, 'balanced'])

# Logistic
P['base logistic'] =\
    {'max_iter': 100,
     'multi_class': 'auto',
     'penalty': 'none',
     'class_weight': None,
     'solver': 'lbfgs'}

# Ridge classifier
P['base ridge'] = {'max_iter': 100,
                   'penalty': 'l2',
                   'solver': 'saga'}

P['ridge C'] =\
    {'max_iter': 100,
     'solver': 'saga',
     'C': Log(lower=1e-5, upper=1e3),
     'class_weight': cls_weight}

P['ridge C extra'] = P['ridge C'].copy()
P['ridge C extra']['max_iter'] =\
    Scalar(lower=100, upper=1000).set_integer_casting()
P['ridge C extra']['tol'] = Log(lower=1e-6, upper=.01)

# Ridge regressor
P['base ridge regressor'] = {'max_iter': 100,
                             'solver': 'lsqr'}

P['ridge regressor dist'] = P['base ridge regressor'].copy()
P['ridge regressor dist']['alpha'] = Log(lower=1e-3, upper=1e5)

# Lasso regressor
P['base lasso regressor'] = {'max_iter': 100}
P['lasso regressor dist'] = P['base lasso regressor'].copy()
P['lasso regressor dist']['alpha'] = Log(lower=1e-5, upper=1e5)

# Lasso classifier
P['base lasso'] = P['base logistic'].copy()
P['base lasso']['solver'] = 'liblinear'
P['base lasso']['penalty'] = 'l1'

P['lasso C'] = P['base lasso'].copy()
P['lasso C']['C'] = Log(lower=1e-5, upper=1e3)
P['lasso C']['class_weight'] = cls_weight

P['lasso C extra'] = P['lasso C'].copy()
P['lasso C extra']['max_iter'] =\
    Scalar(lower=100, upper=1000).set_integer_casting()
P['lasso C extra']['tol'] = Log(lower=1e-6, upper=.01)

# Elastic net classifier
P['base elastic'] = P['base logistic'].copy()
P['base elastic']['penalty'] = 'elasticnet'
P['base elastic']['l1_ratio'] = .5
P['base elastic']['solver'] = 'saga'

P['elastic classifier'] = P['base elastic'].copy()
P['elastic classifier']['C'] = Log(lower=1e-5, upper=1e5)
P['elastic classifier']['l1_ratio'] = Scalar(lower=.01, upper=1)
P['elastic classifier']['class_weight'] = cls_weight

P['elastic clf v2'] = P['elastic classifier'].copy()
P['elastic clf v2']['C'] = Log(lower=1e-2, upper=1e5)

P['elastic classifier extra'] = P['elastic classifier'].copy()
P['elastic classifier extra']['max_iter'] =\
    Scalar(lower=100, upper=1000).set_integer_casting()
P['elastic classifier extra']['tol'] = Log(lower=1e-6, upper=.01)

# Elastic net regression
P['base elastic net'] = {'max_iter': 100}
P['elastic regression'] = P['base elastic net'].copy()
P['elastic regression']['alpha'] = Log(lower=1e-5, upper=1e5)
P['elastic regression']['l1_ratio'] = Scalar(lower=.01, upper=1)

P['elastic regression extra'] = P['elastic regression'].copy()
P['elastic regression extra']['max_iter'] =\
    Scalar(lower=100, upper=1000).set_integer_casting()
P['elastic regression extra']['tol'] = Log(lower=1e-6, upper=.01)

# Other
P['base huber'] = {'epsilon': 1.35}
P['base gnb'] = {'var_smoothing': 1e-9}

P['base knn'] = {'n_neighbors': 5}
P['knn dist'] = {'weights': TransitionChoice(['uniform', 'distance']),
                 'n_neighbors':
                 Scalar(lower=2, upper=25).set_integer_casting()}
P['base knn regression'] = P['base knn'].copy()
P['knn dist regression'] = P['knn dist'].copy()

P['base dt'] = {}
P['dt dist'] = {'max_depth':
                Scalar(lower=1, upper=30).set_integer_casting(),
                'min_samples_split':
                Scalar(lower=2, upper=50).set_integer_casting()}

P['dt classifier dist'] = P['dt dist'].copy()
P['dt classifier dist']['class_weight'] = cls_weight

P['base linear'] = {'fit_intercept': True}

P['base rf'] = {'n_estimators': 100}
P['base rf regressor'] = P['base rf'].copy()

# Forest based
n_estimators = Scalar(init=100, lower=3, upper=500).set_integer_casting()
depths =\
    TransitionChoice(
        [None, Scalar(init=25, lower=2, upper=200).set_integer_casting()])

P['rf dist'] = {'n_estimators': n_estimators,
                'max_depth': depths,
                'max_features': Scalar(lower=.1, upper=1.0),
                'min_samples_split': Scalar(lower=.1, upper=1.0),
                'bootstrap': True}

P['rf classifier dist'] = P['rf dist'].copy()
P['rf classifier dist']['class_weight'] = cls_weight

# Light gbm params
P['base lgbm'] = {'silent': True}


P['lgbm dist1'] =\
    {'silent': True,
     'boosting_type': TransitionChoice(['gbdt', 'dart', 'goss']),
     'n_estimators': n_estimators,
     'num_leaves':
         Scalar(init=20, lower=6, upper=80).set_integer_casting(),
         'min_child_samples':
         Scalar(lower=10, upper=500).set_integer_casting(),
         'min_child_weight':
         Log(lower=1e-5, upper=1e4),
         'subsample':
         Scalar(lower=.3, upper=.95),
         'colsample_bytree':
         Scalar(lower=.3, upper=.95),
         'reg_alpha': TransitionChoice([0, Log(lower=1e-5, upper=1)]),
         'reg_lambda': TransitionChoice([0, Log(lower=1e-5, upper=1)])}

P['lgbm dist2'] =\
    {'silent': True,
     'lambda_l2': 0.001,
     'boosting_type': TransitionChoice(['gbdt', 'dart']),
     'min_child_samples':
         TransitionChoice([1, 5, 7, 10, 15, 20, 35, 50, 100, 200, 500, 1000]),
         'num_leaves':
         TransitionChoice([2, 4, 7, 10, 15, 20, 25, 30, 35,
                           40, 50, 65, 80, 100, 125, 150, 200, 250]),
         'colsample_bytree':
         TransitionChoice([0.7, 0.9, 1.0]),
         'subsample':
         Scalar(lower=.3, upper=1),
         'learning_rate':
         TransitionChoice([0.01, 0.05, 0.1]),
         'n_estimators':
         TransitionChoice([5, 20, 35, 50, 75, 100, 150,
                           200, 350, 500, 750, 1000])}

P['lgbm dist3'] = {'silent': True,
                   'n_estimators': 1000,
                   'early_stopping_rounds': 150,
                   'eval_split': .2,
                   'boosting_type': 'gbdt',
                   'learning_rate':
                   'Log(lower=5e-3, upper=.2, init=.1)',
                   'colsample_bytree':
                   'Scalar(lower=.75, upper=1, init=1)',
                   'min_child_samples':
                   Scalar(lower=2, upper=30, init=20).set_integer_casting(),
                   'num_leaves':
                   Scalar(lower=16, upper=96, init=31).set_integer_casting()}

P['lgbm classifier dist1'] = P['lgbm dist1'].copy()
P['lgbm classifier dist1']['class_weight'] = cls_weight

P['lgbm classifier dist2'] = P['lgbm dist2'].copy()
P['lgbm classifier dist2']['class_weight'] = cls_weight

P['lgbm classifier dist3'] = P['lgbm dist3'].copy()
P['lgbm classifier dist3']['class_weight'] = cls_weight

P['base xgb'] = {'verbosity': 0,
                 'objective': 'reg:squarederror'}

P['base xgb classifier'] = P['base xgb'].copy()
P['base xgb classifier']['objective'] = 'binary:logistic'

P['xgb dist1'] =\
    {'verbosity': 0,
     'objective': 'reg:squarederror',
     'n_estimators': n_estimators,
     'min_child_weight': Log(lower=1e-5, upper=1e4),
     'subsample': Scalar(lower=.3, upper=.95),
     'colsample_bytree': Scalar(lower=.3, upper=.95),
     'reg_alpha': TransitionChoice([0, Log(lower=1e-5, upper=1)]),
     'reg_lambda': TransitionChoice([0, Log(lower=1e-5, upper=1)])}

P['xgb dist2'] =\
    {'verbosity': 0,
     'objective': 'reg:squarederror',
     'max_depth': depths,
     'learning_rate': Scalar(lower=.01, upper=.5),
     'n_estimators': Scalar(lower=3, upper=500).set_integer_casting(),
     'min_child_weight': TransitionChoice([1, 5, 10, 50]),
     'subsample': Scalar(lower=.5, upper=1),
     'colsample_bytree': Scalar(lower=.4, upper=.95)}

P['xgb dist3'] =\
    {'verbosity': 0,
     'objective': 'reg:squarederror',
     'learning_rare': Scalar(lower=.005, upper=.3),
     'min_child_weight': Scalar(lower=.5, upper=10),
     'max_depth': TransitionChoice(np.arange(3, 10)),
     'subsample': Scalar(lower=.5, upper=1),
     'colsample_bytree': Scalar(lower=.5, upper=1),
     'reg_alpha': Log(lower=.00001, upper=1)}

P['xgb classifier dist1'] = P['xgb dist1'].copy()
P['xgb classifier dist1']['objective'] = 'binary:logistic'

P['xgb classifier dist2'] = P['xgb dist2'].copy()
P['xgb classifier dist2']['objective'] = 'binary:logistic'

P['xgb classifier dist3'] = P['xgb dist3'].copy()
P['xgb classifier dist3']['objective'] = 'binary:logistic'

P['base gp regressor'] = {'n_restarts_optimizer': 5,
                          'normalize_y': True}
P['base gp classifier'] = {'n_restarts_optimizer': 5}

# probability = True
P['base svm'] = {'kernel': 'rbf',
                 'gamma': 'scale'}

P['svm dist'] = P['base svm'].copy()
P['svm dist']['C'] = Log(lower=1e-4, upper=1e4)
P['svm dist']['gamma'] = Log(lower=1e-6, upper=1)

P['base svm classifier'] = P['base svm'].copy()
P['base svm classifier']['probability'] = True

P['svm classifier dist'] = P['svm dist'].copy()
P['svm classifier dist']['probability'] = True
P['svm classifier dist']['class_weight'] = cls_weight


# Define different choices for the mlp
P['base mlp'] = {}

sizes = Scalar(init=200, lower=50, upper=400).set_integer_casting()
batch_size = TransitionChoice(['auto', sizes])

P['mlp dist 1 layer'] =\
    {'hidden_layer_sizes':
     Scalar(init=100, lower=2, upper=300).set_integer_casting(),
     'activation':
         TransitionChoice(['identity', 'logistic', 'tanh', 'relu']),
         'alpha': Log(lower=1e-5, upper=1e2),
         'batch_size': batch_size,
         'learning_rate':
         TransitionChoice(['constant', 'invscaling', 'adaptive']),
         'learning_rate_init': Log(lower=1e-5, upper=1e2),
         'max_iter':
         Scalar(init=200, lower=100, upper=1000).set_integer_casting(),
         'beta_1': Scalar(init=.9, lower=.1, upper=.99),
         'beta_2': Scalar(init=.999, lower=.1, upper=.9999)}

P['mlp dist es 1 layer'] = P['mlp dist 1 layer'].copy()
P['mlp dist es 1 layer']['early_stopping'] = True
P['mlp dist es 1 layer']['n_iter_no_change'] =\
    Scalar(lower=5, upper=50).set_integer_casting()

two_layer = Array(init=(100, 100)).set_mutation(sigma=50)
two_layer = two_layer.set_bounds(lower=1, upper=300).set_integer_casting()

P['mlp dist 2 layer'] = P['mlp dist 1 layer'].copy()
P['mlp dist 2 layer']['hidden_layer_sizes'] = two_layer

P['mlp dist es 2 layer'] = P['mlp dist es 1 layer'].copy()
P['mlp dist 2 layer']['hidden_layer_sizes'] = two_layer

three_layer = Array(init=(100, 100, 100)).set_mutation(sigma=50)
three_layer = three_layer.set_bounds(lower=1, upper=300).set_integer_casting()

P['mlp dist 3 layer'] = P['mlp dist 1 layer'].copy()
P['mlp dist 3 layer']['hidden_layer_sizes'] = three_layer

P['mlp dist es 3 layer'] = P['mlp dist es 1 layer'].copy()
P['mlp dist 3 layer']['hidden_layer_sizes'] = three_layer

P['base linear svc'] = {'max_iter': 100}

P['linear svc dist'] = P['base linear svc'].copy()
P['linear svc dist']['C'] = Log(lower=1e-4, upper=1e4)
P['linear svc dist']['class_weight'] = cls_weight

P['base linear svr'] = {'loss': 'epsilon_insensitive',
                        'max_iter': 100}

P['linear svr dist'] = P['base linear svr'].copy()
P['linear svr dist']['C'] = Log(lower=1e-4, upper=1e4)

P['base sgd'] = {'loss': 'squared_loss'}


loss_choice = TransitionChoice(['hinge', 'log', 'modified_huber',
                                'squared_hinge', 'perceptron'])

lr_choice = TransitionChoice(['optimal', 'invscaling',
                              'adaptive', 'constant'])

P['sgd classifier big search'] =\
    {'loss': loss_choice,
     'penalty': TransitionChoice(['l2', 'l1', 'elasticnet']),
     'alpha': Log(lower=1e-5, upper=1e2),
     'l1_ratio': Scalar(lower=.01, upper=1),
     'max_iter': 100,
     'learning_rate': lr_choice,
     'eta0': Log(lower=1e-6, upper=1e3),
     'power_t': Scalar(lower=.1, upper=.9),
     'early_stopping': TransitionChoice([False, True]),
     'validation_fraction': Scalar(lower=.05, upper=.5),
     'n_iter_no_change': Scalar(lower=5, upper=30).set_integer_casting(),
     'class_weight': cls_weight}

# Make elastic net version
P['sgd elastic'] =\
    {'loss': 'squared_epsilon_insensitive',
     'penalty': 'elasticnet',
     'alpha': Log(lower=1e-5, upper=1e5),
     'l1_ratio': Scalar(lower=.01, upper=1)}

P['sgd elastic classifier'] = P['sgd elastic'].copy()
P['sgd elastic classifier']['class_weight'] = cls_weight

# Auto gluon
P['pt binary'] = {'problem_type': 'binary'}
P['pt multiclass'] = {'problem_type': 'multiclass'}
P['pt regression'] = {'problem_type': 'regression'}



from . import (imputers, models, ensembles, feature_selectors,
               loaders, scalers, transformers)

__all__ = ['imputers', 'models', 'ensembles', 'feature_selectors',
           'loaders', 'scalers', 'transformers']


import pandas as pd
import numpy as np
import glob
import os
import warnings
from joblib import wrap_non_picklable_objects


def proc_fop(fop):

    # If provided as just % number, divide by 100
    if not isinstance(fop, tuple):
        fop /= 100
        return (fop, 1-fop)

    elif fop[0] is None:
        return tuple([None, 1-(fop[1] / 100)])

    elif fop[1] is None:
        return tuple([fop[0] / 100, None])

    return tuple([fop[0]/100, 1-(fop[1] / 100)])


def auto_file_to_subject(in_path):

    base_name = os.path.basename(in_path)
    return os.path.splitext(base_name)[0]


def proc_file_input(files, file_to_subject):

    if not isinstance(files, dict):
        raise ValueError('files must be passed as a python dict')

    if file_to_subject == 'auto':
        file_to_subject = auto_file_to_subject

    if file_to_subject is None:
        raise RuntimeError('file_to_subject must be specified!')

    # If not passed as dict, convert to dict
    if not isinstance(file_to_subject, dict):
        file_to_subject = {key: file_to_subject for key in files}

    # Check
    for key in files:
        if key not in file_to_subject:
            raise ValueError('If passing file_to_subject as a dict '
                             'then a value must be passed for all '
                             'keys in files. ' + repr(key) + 'was '
                             'not passed in this case.')

    # Compute pd series
    files_series = dict()
    for key in files:

        file_paths = files[key]

        # If passed file path as str, assume globbing
        if isinstance(file_paths, str):
            file_paths = glob.glob(file_paths)

        subjects = [file_to_subject[key](fp) for fp in file_paths]
        files_series[key] = pd.Series(file_paths, index=subjects,
                                      dtype='object')

    return files_series


def base_load_subjects(subjects):

    loaded_subjects = set()

    if isinstance(subjects, str):
        with open(subjects, 'r') as f:
            lines = f.readlines()

            for line in lines:
                subject = line.rstrip()

                try:
                    subject = eval(subject)
                except NameError:
                    subject = subject

                loaded_subjects.add(subject)

    else:
        loaded_subjects = set([s for s in subjects])

    return loaded_subjects


def save_subjects(loc, subjects):

    with open(loc, 'w') as f:
        for subject in subjects:
            f.write(repr(subject) + '\n')


def add_new_categories(existing, new_values):

    # This is only relevant for type category
    if existing.dtype.name != 'category':
        return

    # Only add new categories
    new_cats = set(pd.unique(new_values))
    existing_cats = set(existing.dtype.categories)
    to_add = new_cats - existing_cats

    # Add in place
    existing.cat.add_categories(list(to_add), inplace=True)


def remove_unused_categories(existing):

    # This is only relevant for type category
    if existing.dtype.name != 'category':
        return

    to_remove = set(existing.dtype.categories) - set(pd.unique(existing))
    to_remove = to_remove - set([np.nan])
    existing.cat.remove_categories(list(to_remove), inplace=True)


def get_str_round(val, places=3):

    if isinstance(val, int):
        return val

    return str(np.round(float(val), places))


def verbose_print(self, *args, **kwargs):
    '''Overriding the print function to allow for
    customizable verbosity.

    According to passed level:

    Warnings are level 0,
    Information on sizes / how many dropped are level 1.
    Set to level -1 to mute warnings too.

    Parameters
    ----------
    args
        Anything that would be passed to default python print
    '''

    if 'level' in kwargs:
        level = kwargs.pop('level')
    else:
        level = 1

    if self.verbose >= level:

        # Use warnings for level = 0
        if level == 0:

            # Conv print to str - then warn
            sep = ' '
            if 'sep' in kwargs:
                sep = kwargs.pop('sep')
            as_str = sep.join(str(arg) for arg in args)

            warnings.warn(as_str)

        # Use base print for rest
        else:
            print(flush=True, *args, **kwargs)


def mp_consol_save(data_files, index, cast_to, save_dr):

    # Load the subjects data
    subj_data = [df.load() for df in data_files]

    # Stack the subj data with extra columns at last axis
    subj_data = np.stack(subj_data, axis=-1)

    # Optional cast to dtype
    if cast_to is not None:
        subj_data = subj_data.astype(cast_to)

    # Save as name of index in save loc
    save_loc = os.path.join(save_dr, str(index) + '.npy')
    np.save(save_loc, subj_data)

    return save_loc


def wrap_load_func(load_func, _print):

    if load_func.__module__ == '__main__':
        wrapped_load_func = wrap_non_picklable_objects(load_func)
        _print('Warning: Passed load_func was defined within the',
               '__main__ namespace and therefore has been '
               'cloud wrapped.',
               'The function will still work, but it is '
               'reccomended to',
               'define this function in a separate file, '
               'and then import',
               'it , otherwise loader caching will be limited',
               'in utility!', level=0)
    else:
        wrapped_load_func = load_func

    return wrapped_load_func



from .FeatSelectors import FeatureSelector
from .loaders import Identity, SurfLabels, SurfMaps

try:
    from .loaders import SingleConnectivityMeasure
except ImportError:
    class SingleConnectivityMeasure():
        pass

try:
    from .loaders import ThresholdNetworkMeasures
except ImportError:
    class ThresholdNetworkMeasures():
        pass

from .MLP import MLPRegressor_Wrapper, MLPClassifier_Wrapper
from .random_parcellation import RandomParcellation
from .Scalers import Winsorizer
from .residualizer import LinearResidualizer


__all__ = ['FeatureSelector',
           'Identity', 'SurfLabels', 'SingleConnectivityMeasure',
           'MLPRegressor_Wrapper',
           'MLPClassifier_Wrapper', 'LinearResidualizer',
           'RandomParcellation', 'Winsorizer',
           'ThresholdNetworkMeasures', 'SurfMaps']



import pytest
from ..BPtPipeline import BPtPipeline
from .helpers import ToFixedTransformer, get_fake_mapping, clean_fake_mapping
from ..ScopeObjs import ScopeTransformer
from ..BPtModel import BPtModel
from ..BPtLoader import BPtLoader
from ...extensions import Identity
from sklearn.linear_model import LinearRegression
import numpy as np
import pandas as pd
import os
import tempfile
from ...default.params.Params import Choice, TransitionChoice
from .helpers import get_param_search
from ..BPtSearchCV import NevergradSearchCV
import shutil
from joblib import hash as joblib_hash
from sklearn.feature_selection import SelectKBest
from ..BPtFeatureSelector import BPtFeatureSelector
from ..BPtTransformer import BPtTransformer
import warnings
from sklearn.decomposition import PCA


def test_BPtPipeline():

    # 'loaders', 'imputers',
    #  'scalers',
    #  'transformers',
    #  'feat_selectors', 'model']

    steps = []
    to_ones = ToFixedTransformer(to=1)
    st = ScopeTransformer(estimator=to_ones, inds=[1, 2])
    steps.append(('to_ones', st))

    model = BPtModel(estimator=LinearRegression(), inds=[0, 1])
    steps.append(('model', model))

    pipe = BPtPipeline(steps=steps)

    X = np.zeros((3, 3))
    y = np.ones(3)

    pipe.fit(X, y)
    assert pipe['to_ones'].inds_ == [1, 2]

    # Should update so that next inds are 0, 2
    # as 1 -> 0, 2 -> 1, 0 -> 2, so 0,1 -> 2, 0, sorted = 0, 2
    assert pipe['model'].inds_ == [0, 2]
    assert len(pipe.mapping_) == 3
    assert pipe.mapping_[0] == 2
    assert pipe.mapping_[1] == 0
    assert pipe.mapping_[2] == 1

    # Make sure when re-fit resets mapping each time
    pipe.fit(X, y)
    assert pipe.mapping_[0] == 2
    assert pipe.mapping_[1] == 0
    assert pipe.mapping_[2] == 1

    # Test propegate n_jobs
    pipe.n_jobs = 2
    assert pipe['to_ones'].n_jobs == 2
    assert pipe['to_ones'].estimator.n_jobs == 2

    X_df = pd.DataFrame(X)
    X_trans = pipe.transform_df(X_df)
    assert X_trans[0].sum() == 0
    assert X_trans[1].sum() == 3
    assert X_trans[2].sum() == 3


def test_skip_loader_no_inds():

    steps = []

    loader = BPtLoader(estimator=Identity(),
                       inds=[],
                       file_mapping={})
    steps.append(('loader', loader))

    to_ones = ToFixedTransformer(to=1)
    st = ScopeTransformer(estimator=to_ones, inds=[1, 2])
    steps.append(('to_ones', st))

    model = BPtModel(estimator=LinearRegression(), inds=[0, 1])
    steps.append(('model', model))

    pipe = BPtPipeline(steps=steps)

    X = np.zeros((3, 3))
    y = np.ones(3)

    # If no errors then means worked
    # since loader isn't constructed correctly.
    pipe.fit(X, y)
    assert pipe.steps[0][1].estimator_ is None


def test_file_mapping_hash():

    # Make sure that regardless of DataFile position
    # in memory, that it hashes correctly.
    mapping = get_fake_mapping(10)
    h1 = joblib_hash(mapping)
    clean_fake_mapping(10)

    mapping = get_fake_mapping(10)
    h2 = joblib_hash(mapping)
    clean_fake_mapping(10)
    assert h1 == h2


def run_pipe_with_loader_ts(cache_loc=None):

    steps = []

    # Loader - transform (5, 2) to (5, 8)
    # as each DataFile contains np.zeros((2, 2))
    file_mapping = get_fake_mapping(100)
    loader = BPtLoader(estimator=Identity(),
                       inds=[0, 1],
                       file_mapping=file_mapping,
                       n_jobs=1,
                       fix_n_jobs=False,
                       cache_loc=None)
    steps.append(('loader', loader))

    # Add transformer to ones
    # input here should be (5, 8) of real val, original
    # inds of 0 should work on half
    to_ones = ToFixedTransformer(to=1)
    st = ScopeTransformer(estimator=to_ones, inds=[0])
    steps.append(('to_ones', st))

    # Add basic linear regression model
    # Original inds should work on all
    model = BPtModel(estimator=LinearRegression(), inds=[0, 1])
    param_dists = {'estimator__fit_intercept': Choice([True, False]),
                   'estimator__normalize':
                   TransitionChoice([True, False])}
    search_model = NevergradSearchCV(estimator=model,
                                     ps=get_param_search(),
                                     param_distributions=param_dists)

    steps.append(('model', search_model))

    # Create pipe
    pipe = BPtPipeline(steps=steps,
                       cache_loc=cache_loc)

    X = np.arange(100).reshape((50, 2))
    y = np.ones(50)

    pipe.fit(X, y, fit_index=np.arange(50))

    # Make sure fit worked correctly
    assert pipe[0].n_features_in_ == 2
    assert pipe[1].n_features_in_ == 8
    assert pipe[1].estimator_.n_features_in_ == 4
    assert len(pipe.mapping_[0]) == 4
    assert len(pipe.mapping_[1]) == 4
    assert 7 in pipe.mapping_[1]

    # Make sure reverse transform works
    X_df = pd.DataFrame(X)

    X_trans = pipe.transform_df(X_df)
    assert X_trans.shape == (50, 8)
    assert X_trans.loc[4, '1_3'] == 9
    assert X_trans.loc[1, '1_2'] == 3
    assert X_trans.loc[4, '0_0'] == 1
    assert X_trans.loc[0, '0_0'] == 1

    # Make sure predict works,
    # seems safe to assume model
    # can learn to predict 1's
    # as all targets are 1's.
    # but may need to change?
    preds = pipe.predict(X)
    assert np.all(preds > .99)

    # Check bpt pipeline coef attribute
    assert np.array_equal(pipe[-1].best_estimator_.coef_,
                          pipe.coef_)

    # Clean fake file mapping
    clean_fake_mapping(100)

    return pipe


def test_pipeline_with_loader():

    # Base pipeline with loader tests
    run_pipe_with_loader_ts(cache_loc=None)


def test_pipeline_fit_caching():

    # Run with cache fit dr
    cache_loc =\
        os.path.join(tempfile.gettempdir(), 'test_cache')

    # If already exists, say from a failed test
    # delete
    if os.path.exists(cache_loc):
        shutil.rmtree(cache_loc)

    pipe = run_pipe_with_loader_ts(cache_loc=cache_loc)

    # Make sure computed hash + saved copy
    assert hasattr(pipe, 'hash_',)
    assert os.path.exists(pipe._get_hash_loc())

    # Delete existing pipe
    del pipe

    # Run again a few times to make sure loading from cache works
    for i in range(5):
        pipe = run_pipe_with_loader_ts(cache_loc=cache_loc)
        assert hasattr(pipe, 'hash_')
        assert pipe.loaded_ is True
        del pipe

    # Removed cached once done
    shutil.rmtree(cache_loc)


def test_pipeline_inverse_transform_FIs_loader_static_transform():

    steps = []

    # Loader - transform (5, 2) to (5, 8)
    # as each DataFile contains np.zeros((2, 2))

    loader = BPtLoader(estimator=Identity(),
                       inds=[1],
                       file_mapping=get_fake_mapping(100))
    steps.append(('loader', loader))

    # Add transformer to ones
    # input here should be (5, 8) of real val, original
    # inds of 0 should work on half
    to_ones = ToFixedTransformer(to=1)
    st = ScopeTransformer(estimator=to_ones, inds=Ellipsis)
    steps.append(('to_ones', st))

    # Add basic linear regression model
    # Original inds should work on all
    model = BPtModel(estimator=LinearRegression(), inds=Ellipsis)
    steps.append(('model', model))

    # Create pipe
    pipe = BPtPipeline(steps=steps)

    X = pd.DataFrame(np.arange(100).reshape((50, 2)))
    y = np.ones(50)

    pipe.fit(X, y)

    # Fake coef
    coef_ = [0, 1, 2, 3, 4]
    feat_names = pipe.transform_feat_names(X)

    fis = pd.Series(coef_, index=feat_names)
    inverse_fis = pipe.inverse_transform_FIs(fis)
    print(inverse_fis)

    assert inverse_fis.loc[0] == 4
    assert inverse_fis.loc[1].shape == ((2, 2))
    assert inverse_fis.loc[1][0][0] == 0

    clean_fake_mapping(100)


def test_pipeline_inverse_transform_FIs_loader_fs():

    warnings.filterwarnings("ignore")

    steps = []

    # Loader - transform (5, 2) to (5, 8)
    # as each DataFile contains np.zeros((2, 2))

    loader = BPtLoader(estimator=Identity(),
                       inds=[0, 1],
                       file_mapping=get_fake_mapping(100))
    steps.append(('loader', loader))

    kbest = SelectKBest(k=3)
    fs = BPtFeatureSelector(kbest, Ellipsis)
    steps.append(('kbest', fs))

    # Add basic linear regression model
    # Original inds should work on all
    model = BPtModel(estimator=LinearRegression(), inds=Ellipsis)
    steps.append(('model', model))

    # Create pipe
    pipe = BPtPipeline(steps=steps)

    X = pd.DataFrame(np.arange(100).reshape((50, 2)))
    y = np.arange(50)

    pipe.fit(X, y)

    coef_ = pipe.coef_

    feat_names = pipe.transform_feat_names(X)
    fis = pd.Series(coef_, index=feat_names)
    inverse_fis = pipe.inverse_transform_FIs(fis)

    # Don't make assumptions on specific coef
    assert inverse_fis.loc[0].shape == (2, 2)
    assert inverse_fis.loc[1].shape == (2, 2)
    assert np.sum(np.sum(inverse_fis)) == np.sum(coef_)

    # Clean up
    clean_fake_mapping(100)


def test_pipeline_inverse_transform_FIs_loader_fs_v2():

    warnings.filterwarnings("ignore")

    steps = []

    # Loader - transform (5, 2) to (5, 8)
    # as each DataFile contains np.zeros((2, 2))

    loader = BPtLoader(estimator=Identity(),
                       inds=[0, 1],
                       file_mapping=get_fake_mapping(100))
    steps.append(('loader', loader))

    to_ones = ToFixedTransformer(to=.15)
    st = ScopeTransformer(estimator=to_ones, inds=[0, 1])
    steps.append(('to_ones', st))

    kbest = SelectKBest(k=3)
    fs = BPtFeatureSelector(kbest, Ellipsis)
    steps.append(('kbest', fs))

    # Add basic linear regression model
    # Original inds should work on all
    model = BPtModel(estimator=LinearRegression(), inds=Ellipsis)
    steps.append(('model', model))

    # Create pipe
    pipe = BPtPipeline(steps=steps)

    X = pd.DataFrame(np.arange(100).reshape((50, 2)))
    y = np.arange(50)

    pipe.fit(X, y)

    coef_ = pipe.coef_

    feat_names = pipe.transform_feat_names(X)
    fis = pd.Series(coef_, index=feat_names)
    inverse_fis = pipe.inverse_transform_FIs(fis)

    # Don't make assumptions on specific coef
    assert inverse_fis.loc[0].shape == (2, 2)
    assert inverse_fis.loc[1].shape == (2, 2)
    assert np.sum(np.sum(inverse_fis)) == np.sum(coef_)

    # Clean up
    clean_fake_mapping(100)


def test_pipeline_inverse_transform_FIs_loader_fs_v3():

    warnings.filterwarnings("ignore")

    steps = []

    # Loader - transform (5, 2) to (5, 8)
    # as each DataFile contains np.zeros((2, 2))

    loader = BPtLoader(estimator=Identity(),
                       inds=[0, 1],
                       file_mapping=get_fake_mapping(100))
    steps.append(('loader', loader))

    to_ones = ToFixedTransformer(to=.15)
    st = ScopeTransformer(estimator=to_ones, inds=[0])
    steps.append(('to_ones', st))

    kbest = SelectKBest(k=3)
    fs = BPtFeatureSelector(kbest, Ellipsis)
    steps.append(('kbest', fs))

    # Add basic linear regression model
    # Original inds should work on all
    model = BPtModel(estimator=LinearRegression(), inds=Ellipsis)
    steps.append(('model', model))

    # Create pipe
    pipe = BPtPipeline(steps=steps)

    X = pd.DataFrame(np.arange(100).reshape((50, 2)))
    y = np.arange(50)

    pipe.fit(X, y)

    coef_ = pipe.coef_

    feat_names = pipe.transform_feat_names(X)
    fis = pd.Series(coef_, index=feat_names)
    inverse_fis = pipe.inverse_transform_FIs(fis)

    # Don't make assumptions on specific coef
    assert inverse_fis.loc[0].shape == (2, 2)
    assert inverse_fis.loc[1].shape == (2, 2)
    assert np.sum(np.sum(inverse_fis)) == np.sum(coef_)

    # Clean up
    clean_fake_mapping(100)


def test_pipeline_inverse_transform_FIs_impossible():

    warnings.filterwarnings("ignore")

    steps = []

    # Loader - transform (5, 2) to (5, 8)
    # as each DataFile contains np.zeros((2, 2))

    loader = BPtLoader(estimator=Identity(),
                       inds=[0, 1],
                       file_mapping=get_fake_mapping(100))
    steps.append(('loader', loader))

    pca = PCA(n_components=3)
    st = BPtTransformer(estimator=pca, inds=[0, 1])
    steps.append(('pca', st))

    # Add basic linear regression model
    # Original inds should work on all
    model = BPtModel(estimator=LinearRegression(), inds=Ellipsis)
    steps.append(('model', model))

    # Create pipe
    pipe = BPtPipeline(steps=steps)

    X = pd.DataFrame(np.arange(100).reshape((50, 2)))
    y = np.arange(50)

    pipe.fit(X, y)

    coef_ = pipe.coef_

    feat_names = pipe.transform_feat_names(X)

    fis = pd.Series(coef_, index=feat_names)

    with pytest.raises(IndexError):
        pipe.inverse_transform_FIs(fis)

    # Clean up
    clean_fake_mapping(100)
